# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
class utils():

#     @staticmethod
#     def regression(phase, topo, fit_intercept=True):
#         import numpy as np
#         import xarray as xr
#         from sklearn.linear_model import LinearRegression
#         from sklearn.pipeline import make_pipeline
#         from sklearn.preprocessing import StandardScaler
# 
#         # define on the same grid
#         topo = topo.reindex_like(phase, method='nearest')
#     
#         # build prediction model with or without plane removal (fit_intercept)
#         regr = make_pipeline(StandardScaler(), LinearRegression(fit_intercept=fit_intercept))
# 
#         topo_values = topo.values.ravel()
#         phase_values = phase.values.ravel()
#         nanmask = np.isnan(a) | np.isnan(b)
# 
#         # fit on non-NaN values only and predict on the full grid
#         phase_topo = regr.fit(np.column_stack([topo_values[~nanmask]]),
#                               np.column_stack([phase_values[~nanmask]]))\
#                     .predict(np.column_stack([topo_values])).reshape(phase.shape)
#         return xr.DataArray(phase_topo, coords=phase.coords)

    @staticmethod
    def histogram(data, bins, range):
        """
        hist, bins = utils.histogram(corr60m.mean('pair'), 10, (0,1))
        print ('bins', bins)
        hist = dask.compute(hist)[0]
        print (hist)
        """
        import dask
        return dask.array.histogram(data, bins=bins, range=range)

    @staticmethod
    def corrcoef(data, mask_diagonal=False):
        """
        Calculate the correlation coefficients matrix for a stack of interferograms.

        Parameters:
        - data (xarray.DataArray): A 3D input grid representing a stack of interferograms.
                                  The dimensions should include time, 'y', and 'x'.
        - mask_diagonal (bool): If True, the diagonal elements of the correlation matrix
                                will be set to NaN to exclude self-correlation.

        Returns:
        xarray.DataArray: The cross-correlation matrix.
        
        Example:
        # plot variance-covariance matrix
        corr_stack = corr60m.mean('pair')
        corr = utils.corrcoef(unwrap.phase.where(corr_stack.where(corr_stack>0.7)))
        corr.where((corr)>0.7).plot.imshow(vmin=-1, vmax=1)
        plt.xticks(rotation=90)
        plt.show()
        
        # find the most correlated interferograms to exclude
        corr = utils.corrcoef(unwrap.phase.where(corr_stack.where(corr_stack>0.7)), mask_diagonal=True)
        df = corr.where(corr>0.7).to_dataframe('').reset_index().dropna()
        exclude = np.unique(np.concatenate([df.ref, df.rep]))
        exclude
        """
        import xarray as xr
        import dask
        import numpy as np
        assert len(data.dims) == 3, 'ERROR: expected 3D input grid'
        stackdim = data.dims[0]
        stackvals = data[stackdim].values
        corr = dask.array.corrcoef(data.stack(points=['y', 'x']).dropna(dim='points').data).round(2)
        corr = xr.DataArray((corr), coords={'ref': stackvals, 'rep': stackvals}).rename('corr')
        corr['ref'] = corr['ref'].astype(str)
        corr['rep'] = corr['rep'].astype(str)
        if mask_diagonal:
            # set diagonal elements to NaN
            diagonal_mask = np.eye(len(stackvals), dtype=bool)
            return corr.where(~diagonal_mask, np.nan)
        return corr

    @staticmethod
    def binary_erosion(data, *args, **kwargs):
        """
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.binary_erosion.html
        """
        import xarray as xr
        from scipy.ndimage import binary_erosion
        array = binary_erosion(data.values, *args, **kwargs)
        return xr.DataArray(array, coords=data.coords, dims=data.dims, attrs=data.attrs)

    @staticmethod
    def binary_dilation(data, *args, **kwargs):
        """
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.binary_dilation.html
        """
        import xarray as xr
        from scipy.ndimage import binary_dilation
        array = binary_dilation(data.values, *args, **kwargs)
        return xr.DataArray(array, coords=data.coords, dims=data.dims, attrs=data.attrs)

    @staticmethod
    def binary_opening(data, *args, **kwargs):
        """
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.binary_opening.html
        
        corrmask = utils.binary_closing(corrmask, structure=np.ones((10,10)))
        corrmask = utils.binary_opening(corrmask, structure=np.ones((10,10)))
        """
        import xarray as xr
        from scipy.ndimage import binary_opening
        array = binary_opening(data.values, *args, **kwargs)
        return xr.DataArray(array, coords=data.coords, dims=data.dims, attrs=data.attrs)

    @staticmethod
    def binary_closing(data, *args, **kwargs):
        """
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.binary_opening.html
        
        corrmask = utils.binary_closing(corrmask, structure=np.ones((10,10)))
        corrmask = utils.binary_opening(corrmask, structure=np.ones((10,10)))
        """
        import xarray as xr
        from scipy.ndimage import binary_closing
        array = binary_closing(data.values, *args, **kwargs)
        return xr.DataArray(array, coords=data.coords, dims=data.dims, attrs=data.attrs)



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2022, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_landmask import Stack_landmask

class Stack_unwrap_snaphu(Stack_landmask):

    # -s for SMOOTH mode and -d for DEFO mode when DEFOMAX_CYCLE should be defined in the configuration
    # DEFO mode (-d) and DEFOMAX_CYCLE=0 is equal to SMOOTH mode (-s)
    # https://web.stanford.edu/group/radar/softwareandlinks/sw/snaphu/snaphu_man1.html
    def snaphu(self, phase, corr=None, conf=None, conncomp=False, debug=False):
        """
        Unwraps phase using SNAPHU with the given phase and correlation data.

        This function unwraps the phase of an interferogram using the Statistical-cost, Network-flow Algorithm
        for Phase Unwrapping (SNAPHU) with user-defined parameters. The unwrapped phase is saved as a grid file
        in the working directory.

        Parameters
        ----------
        phase : xarray.DataArray
            The phase data as a string or xarray.DataArray, default is 'phasefilt'.

        corr : xarray.DataArray, optional
            The correlation data as a string or xarray.DataArray, default is 'corr'.

        conf : str, optional
            The SNAPHU configuration string, default is None (use the PRM's snaphu_config method).

        conncomp : bool, optional
            If True, return connection components map, default is False.

        debug : bool, optional
            If True, print debugging information during the unwrapping process, default is False.

        Returns
        -------
        xarray.Dataset
            Return the unwrapped phase and optional connection components as an xarray.Dataset.

        """
        import xarray as xr
        import numpy as np
        import pandas as pd
        import os
        import subprocess
        from datetime import datetime
        import uuid
        # disable "distributed.utils_perf - WARNING - full garbage collections ..."
        from dask.distributed import utils_perf
        utils_perf.disable_gc_diagnosis()

        if conf is None:
            conf = self.snaphu_config()
        # set unique processing subdirectory
        conf += f'    TILEDIR snaphu_tiledir_{str(uuid.uuid4())}'

        # define basename for SNAPHU temp files
        # crop .grd from filename
        basename = self.get_filename(f'snaphu_{str(uuid.uuid4())}', '')[:-4]
        #print ('basename', basename)

        # SNAPHU input files
        phase_in = basename + '.phase'
        corr_in = basename + '.corr'
        mask_in = basename + '.bytemask'
        # SNAPHU output files
        unwrap_out = basename + 'unwrap.out'
        conncomp_out = basename + 'conncomp.out'

        # prepare SNAPHU input files
        # NaN values are not allowed for SNAPHU phase input file
        # interpolate when exist valid values around and fill zero pixels far away from valid ones
        # convert to lazy array setting the chunk size
        phase.fillna(0).compute(n_workers=1).values.astype(np.float32).tofile(phase_in)
        # SNAPHU masks out 0 and uses only valid pixels with mask 1
        xr.where(np.isnan(phase), 0, 1).values.astype(np.ubyte).tofile(mask_in)
    
        if corr is not None:
            # NaN values are not allowed for SNAPHU correlation input file
            # just fill NaNs by zeroes because the main trick is phase filling
            corr.fillna(0).compute(n_workers=1).values.astype(np.float32).tofile(corr_in)

        # launch SNAPHU binary (NaNs are not allowed for input but returned in output)
        argv = ['snaphu', phase_in, str(phase.shape[1]), '-M', mask_in, '-f', '/dev/stdin', '-o', unwrap_out, '-d']
        # output connection componetets map
        if conncomp:
            argv.append('-g')
            argv.append(conncomp_out)
        # add optional correlation grid
        if corr is not None:
            argv.append('-c')
            argv.append(corr_in)
        if debug:
            argv.append('-v')
            print ('DEBUG: argv', argv)
        p = subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             encoding='utf8', bufsize=10*1000*1000)
        stdout_data, stderr_data = p.communicate(input=conf)

        outs = []
        # check for expected SNAPHU output files
        if os.path.exists(unwrap_out) and (not conncomp or os.path.exists(conncomp_out)):
            # convert to grid unwrapped phase from SNAPHU output applying postprocessing
            values = np.fromfile(unwrap_out, dtype=np.float32).reshape(phase.shape)
            #values = np.frombuffer(stdout_data, dtype=np.float32).reshape(phase.shape)
            # revert NaNs in output because SNAPNU does not support them
            unwrap = xr.DataArray(values, phase.coords, name='phase').chunk(self.chunksize).where(~np.isnan(phase))
            outs.append(unwrap)
            del values, unwrap

            if conncomp:
                # convert to grid the connected components from SNAPHU output as is (UCHAR)
                values = np.fromfile(conncomp_out, dtype=np.ubyte).reshape(phase.shape)
                conn = xr.DataArray(values, phase.coords, name='conncomp').chunk(self.chunksize)
                outs.append(conn)
                del values, conn
        else:
            # return the same data structure as expected but NaN-filled
            outs.append(xr.full_like(phase, np.nan).rename('phase'))
            if conncomp:
                outs.append(xr.full_like(phase, np.nan).rename('conncomp'))
        out = xr.merge(outs)
        del outs

        # add processing log
        out.attrs['snaphu'] = stdout_data + '\n' + stderr_data

        # the output files deleted immediately
        # but these are accessible while open descriptors persist
        for tmp_file in [phase_in, corr_in, mask_in, unwrap_out, conncomp_out]:
            if os.path.exists(tmp_file):
                os.remove(tmp_file)

        return out

    def snaphu_config(self, defomax=0, **kwargs):
        """
        Generate a Snaphu configuration file.

        Parameters
        ----------
        defomax : int, optional
            Maximum deformation value. Default is 0.
        **kwargs : dict, optional
            Additional parameters to include in the configuration file.

        Returns
        -------
        str
            The Snaphu configuration file content.

        Notes
        -----
        This method uses the `snaphu_config` method of the PRM object.

        Examples
        --------
        Generate a Snaphu configuration file with defomax=10:
        snaphu_config(defomax=10)

        Generate a Snaphu configuration file with defomax=5 and additional parameters:
        snaphu_config(defomax=5, param1=10, param2=20)
        """
        #return self.PRM().snaphu_config(defomax, **kwargs)
        import os
        import joblib

        tiledir = os.path.splitext(self.PRM().filename)[0]
        n_jobs = joblib.cpu_count()

        conf_basic = f"""
        # basic config
        INFILEFORMAT   FLOAT_DATA
        OUTFILEFORMAT  FLOAT_DATA
        AMPFILEFORMAT  FLOAT_DATA
        CORRFILEFORMAT FLOAT_DATA
        ALTITUDE       693000.0
        EARTHRADIUS    6378000.0
        NEARRANGE      831000
        DR             18.4
        DA             28.2
        RANGERES       28
        AZRES          44
        LAMBDA         0.0554658
        NLOOKSRANGE    1
        NLOOKSAZ       1
        TILEDIR        {tiledir}_snaphu_tiledir
        NPROC          {n_jobs}
        """
        conf_custom = '# custom config\n'
        # defomax can be None
        keyvalues = ([('DEFOMAX_CYCLE', defomax)] if defomax is not None else []) + list(kwargs.items())
        for key, value in keyvalues:
            if isinstance(value, bool):
                value = 'TRUE' if value else 'FALSE'
            conf_custom += f'        {key} {value}\n'
        return conf_basic + conf_custom



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_reframe import Stack_reframe
from .PRM import PRM

class Stack_dem(Stack_reframe):

    # Buffer size in degrees to expand the area covered by the DEM
    buffer_degrees = 0.1

    def get_extent_ra(self):
        """
        minx, miny, maxx, maxy = np.round(geom.bounds).astype(int)
        """
        import numpy as np
        from shapely.geometry import LineString

        dem = self.get_dem()
        df = dem.isel(lon=[0,-1]).to_dataframe().reset_index()
        geom = self.geocode(LineString(np.column_stack([df.lon, df.lat])))
        return geom

    def get_extent(self, grid=None, subswath=None):
        import numpy as np

        extent = self.get_reference(subswath).dissolve().envelope.item()
        if grid is None:
            return extent
        bounds = np.round(extent.bounds, 3)
        #print ('xmin, xmax', xmin, xmax)
        return grid\
               .transpose('lat','lon')\
               .sel(lat=slice(bounds[1] - self.buffer_degrees, bounds[3] + self.buffer_degrees),
                    lon=slice(bounds[0] - self.buffer_degrees, bounds[2] + self.buffer_degrees))

    def get_geoid(self, grid=None):
        """
        Get EGM96 geoid heights.

        Parameters
        ----------
        grid : xarray array or dataset, optional
            Interpolate geoid heights on the grid. Default is None.

        Returns
        -------
        None

        Examples
        --------
        stack.get_geoid()

        Notes
        -----
        See EGM96 geoid heights on http://icgem.gfz-potsdam.de/tom_longtime
        """
        import xarray as xr
        import os

        gmtsar_sharedir = PRM().gmtsar_sharedir()
        geoid_filename = os.path.join(gmtsar_sharedir, 'geoid_egm96_icgem.grd')
        geoid = xr.open_dataarray(geoid_filename, engine=self.netcdf_engine, chunks=self.netcdf_chunksize).rename({'y': 'lat', 'x': 'lon'})
        if grid is not None:
            geoid = geoid.interp_like(grid, method='cubic')
        return geoid

    def set_dem(self, dem_filename):
        """
        Set the filename of the digital elevation model (DEM) in WGS84 NetCDF grid.

        Parameters
        ----------
        dem_filename : str or None
            Path to the WGS84 NetCDF DEM file. If provided, the DEM filename will be set. If None, the DEM filename will be cleared.

        Returns
        -------
        self
            Returns the modified instance of the class.

        Examples
        --------
        Set the DEM filename:
        stack = stack.set_dem('data/DEM_WGS84.nc')

        Notes
        -----
        This method sets the filename of the digital elevation model (DEM) data to be used in the SAR processing.
        The DEM file is a NetCDF dataset in geographical coordinates containing elevation values for the Earth's surface.
        By setting the DEM filename, it allows SAR processing algorithms to utilize the DEM data for geocoding, topographic
        correction, and other applications. If `dem_filename` is None, the DEM filename will be cleared.
        """
        import os
        if dem_filename is not None:
            assert os.path.exists(dem_filename), f'DEM file not found: {dem_filename}'
            assert os.path.isfile(dem_filename) and os.access(dem_filename, os.R_OK), f'DEM file is not readable: {dem_filename}'
            self.dem_filename = os.path.relpath(dem_filename,'.')
        else:
            self.dem_filename = None
        return self

    # buffer required to get correct (binary) results from SAT_llt2rat tool
    # small buffer produces incomplete area coverage and restricted NaNs
    # 0.02 degrees works well worldwide but not in Siberia
    # minimum buffer size: 8 arc seconds for 90 m DEM
    # subswath argument is required for aligning
    def get_dem(self, subswath=None):
        """
        Retrieve the digital elevation model (DEM) data.

        Parameters
        ----------
        subswath : str, optional
            Subswath name. Default is None.

        Returns
        -------
        xarray.DataArray
            The DEM data as a DataArray object.

        Raises
        ------
        Exception
            If the DEM is not set before calling this method.

        Examples
        --------
        Get DEM for all the processed subswaths:
        topo_ll = stack.get_dem()

        Get DEM for a single subswath IW1:
        topo_ll = stack.get_dem(1)

        Notes
        -----
        This method retrieves the digital elevation model (DEM) data previously downloaded and stored in a NetCDF file.
        The DEM file is opened, and the elevation variable is extracted. Any missing values in the elevation data are filled
        with zeros (mostly representing water surfaces).
        """
        import xarray as xr
        import os
        import warnings
        # supress warnings "UserWarning: The specified chunks separate the stored chunks along dimension"
        warnings.filterwarnings('ignore')

        if self.dem_filename is None:
            raise Exception('Set DEM first')

        # open DEM file and find the elevation variable
        # because sometimes grid includes 'crs' or other variables
        dem = xr.open_dataset(self.dem_filename, engine=self.netcdf_engine, chunks=self.chunksize)
        if 'lat' not in dem.coords and 'y' in dem.coords:
            dem = dem.rename({'y': 'lat', 'x': 'lon'})
        # define latlon array
        z_array_name = [data_var for data_var in dem.data_vars if len(dem.data_vars[data_var].coords)==2]
        assert len(z_array_name) == 1
        # extract the array and fill missed values (mostly water surfaces)
        dem = dem[z_array_name[0]].fillna(0)
        # round the coordinates up to 1 mm
        dem['lat'] = dem.lat.round(8)
        dem['lon'] = dem.lon.round(8)

        return self.get_extent(dem, subswath)

    # buffer required to get correct (binary) results from SAT_llt2rat tool
    # 0.02 degrees works well worldwide but not in Siberia
    # small margin produces insufficient DEM not covers the defined area
    # https://docs.generic-mapping-tools.org/6.0/datasets/earth_relief.html
    # only bicubic interpolation supported as the best one for the case
    def download_dem(self, geometry='auto', product='1s'):
        """
        Download and preprocess digital elevation model (DEM) data.

        Parameters
        ----------
        product : str, optional
            Product type of the DEM data. Available options are '1s' or 'SRTM1' (1 arcsec ~= 30m, default)
            and '3s' or 'SRTM3' (3 arcsec ~= 90m).

        Returns
        -------
        None

        Examples
        --------
        Download default STRM1 DEM (~30 meters):
        stack.download_dem()

        Download STRM3 DEM (~90 meters):
        stack.download_dem(product='STRM3')

        Notes
        -----
        This method uses the GMT servers to download SRTM 1 or 3 arc-second DEM data. The downloaded data is then
        preprocessed by removing the EGM96 geoid to make the heights relative to the WGS84 ellipsoid.
        """
        import geopandas as gpd
        import xarray as xr
        import numpy as np
        import pygmt
        # suppress warnings
        pygmt.config(GMT_VERBOSE='errors')
        import rioxarray as rio
        import os
        #import subprocess
        from tqdm.auto import tqdm
        import warnings
        warnings.filterwarnings('ignore')

        dem_filename = os.path.join(self.basedir, 'DEM_WGS84.nc')

        if self.dem_filename is not None:
            print ('NOTE: DEM exists, ignore the command. Use Stack.set_dem(None) to allow new DEM downloading')
            return

        if product in ['SRTM1', '1s', '01s']:
            resolution = '01s'
        elif product in ['SRTM3', '3s', '03s']:
            resolution = '03s'
        else:
            print (f'ERROR: unknown product {product}. Available only SRTM1 ("01s") and SRTM3 ("03s") DEM using GMT servers')
            return

        # round the coordinates up to 1m
        if geometry is None or (type(geometry) == str and geometry == 'auto'):
            # apply scenes geometry
            geometry = self.get_extent().buffer(self.buffer_degrees)
        elif isinstance(geometry, gpd.GeoDataFrame):
            geometry = geometry.dissolve().envelope.item()
        elif isinstance(geometry, gpd.GeoSeries):
            geometry = geometry.unary_union.envelope
        minx, miny, maxx, maxy = np.round(geometry.bounds, 5)
        #print ('minx, miny, maxx, maxy', minx, miny, maxx, maxy)

        with tqdm(desc='DEM Downloading', total=1) as pbar:
            # download DEM using GMT extent W E S N
            ortho = pygmt.datasets.load_earth_relief(resolution=resolution, region=[minx, maxx, miny, maxy])
            # heights correction
            geoid = self.get_geoid(ortho)
            if os.path.exists(dem_filename):
                os.remove(dem_filename)
            encoding = {'dem': self._compression(ortho.shape)}
            (ortho + geoid).rename('dem').load().to_netcdf(dem_filename, encoding=encoding, engine=self.netcdf_engine)
            pbar.update(1)

        self.dem_filename = dem_filename

    def load_dem(self, filename, geometry='auto'):
        """
        Load and preprocess digital elevation model (DEM) data from specified datafile.

        Parameters
        ----------
        filename : str, optional
            DEM filename.

        Returns
        -------
        None

        Examples
        --------
        Load and crop from local NetCDF file:
        stack.load_dem('GEBCO_2020/GEBCO_2020.nc')

        Load and crop from local GeoTIF file:
        stack.load_dem('GEBCO_2019.tif')

        Notes
        -----
        This method loads DEM from the user specified file. The downloaded data is then preprocessed by removing
        the EGM96 geoid to make the heights relative to the WGS84 ellipsoid.
        """
        import xarray as xr
        import numpy as np
        import rioxarray as rio
        import geopandas as gpd
        import os

        dem_filename = os.path.join(self.basedir, 'DEM_WGS84.nc')

        if self.dem_filename is not None:
            print ('NOTE: DEM exists, ignore the command. Use Stack.set_dem(None) to allow new DEM downloading')
            return

        if os.path.splitext(filename)[-1] in ['.tiff', '.tif', '.TIF']:
            ortho = rio.open_rasterio(filename, chunks=self.chunksize).squeeze(drop=True)\
                .rename({'y': 'lat', 'x': 'lon'})\
                .drop('spatial_ref')
            if ortho.lat.diff('lat')[0].item() < 0:
                ortho = ortho.reindex(lat=ortho.lat[::-1])
        elif os.path.splitext(filename)[-1] in ['.nc', '.netcdf', '.grd']:
            ortho = xr.open_dataarray(filename, engine=self.netcdf_engine, chunks=self.chunksize)
        else:
            print ('ERROR: filename extension is not recognized. Should be one from .tiff, .tif, .TIF, .nc, .netcdf, .grd')

        # crop
        #ortho = self.get_extent(ortho)
        # round the coordinates up to 1m
        if type(geometry) == str and geometry == 'auto':
            # apply scenes geometry
            extent = self.get_extent().buffer(self.buffer_degrees)
        elif isinstance(geometry, gpd.GeoDataFrame):
            extent = geometry.dissolve().envelope.item()
        #minx, miny, maxx, maxy = np.round(geometry.bounds, 5)
        #print ('minx, miny, maxx, maxy', minx, miny, maxx, maxy)
        bounds = np.round(extent.bounds, 5)
        #print ('xmin, xmax', xmin, xmax)
        ortho = ortho\
               .transpose('lat','lon')\
               .sel(lat=slice(bounds[1], bounds[3]),
                    lon=slice(bounds[0], bounds[2]))

        # heights correction
        geoid = self.get_geoid(ortho)
        if os.path.exists(dem_filename):
            os.remove(dem_filename)
        encoding = {'dem': self._compression(ortho.shape)}
        (ortho + geoid).rename('dem').load()\
            .to_netcdf(dem_filename, encoding=encoding, engine=self.netcdf_engine)

        self.dem_filename = dem_filename



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_multilooking import Stack_multilooking

class Stack_landmask(Stack_multilooking):

    def set_landmask(self, landmask_filename):
        """
        Set the landmask file path.

        Parameters
        ----------
        landmask_filename : str or None
            The NetCDF file path of the landmask. If None, the landmask file path will be set to None.

        Examples
        --------
        stack = stack.set_landmask('data/landmask.nc')

        Returns
        -------
        self : Stack
            The Stack object with the updated landmask file path.
        """
        import os
        if landmask_filename is not None:
            assert os.path.exists(landmask_filename), f'Landmask file not found: {landmask_filename}'
            assert os.path.isfile(landmask_filename) and os.access(landmask_filename, os.R_OK), f'Landmask file is not readable: {landmask_filename}'
            self.landmask_filename = os.path.relpath(landmask_filename,'.')
        else:
            self.landmask_filename = None
        return self

    def get_landmask(self):
        """
        Get the landmask grid in geographic coordinates.

        Parameters
        ----------
        None

        Returns
        -------
        landmask : xarray.DataArray
            The landmask grid.

        Examples
        --------
        Get land mask in geographic coordinates:
        landmask = stack.get_landmask()

        Notes
        -----
        This method opens the landmask NetCDF file and extracts the landmask grid.
        """
        import xarray as xr
        import os

        if self.landmask_filename is None:
            raise Exception('Set landmask first')

        # open DEM file and find the elevation variable
        # because sometimes grid includes 'crs' or other variables
        landmask = xr.open_dataset(self.landmask_filename, engine=self.netcdf_engine, chunks=self.chunksize)
        assert 'lat' in landmask.coords and 'lon' in landmask.coords
        # define latlon array
        z_array_name = [data_var for data_var in landmask.data_vars if len(landmask.data_vars[data_var].coords)==2]
        assert len(z_array_name) == 1
        # extract the array and fill missed values by zero (mostly ocean area)
        landmask = landmask[z_array_name[0]].fillna(0)
        # round the coordinates up to 1 mm to have the same grid for dem and landmask
        landmask['lat'] = landmask.lat.round(8)
        landmask['lon'] = landmask.lon.round(8)
        return landmask

    def download_landmask(self, product='1s', debug=False):
        """
        Download the landmask and save as NetCDF file.

        Parameters
        ----------
        product : str, optional
                Available options are '1s' (1 arcsec ~= 30m, default) and '3s' (3 arcsec ~= 90m).
        debug : bool, optional
            Whether to enable debug mode. Default is False.

        Examples
        --------
        stack.download_landmask()

        Notes
        -----
        This method downloads the landmask using GMT's local data or server. The landmask is built based on the interferogram DEM area.
        """
        import pygmt
        import os
        from tqdm.auto import tqdm

        if self.landmask_filename is not None:
            print ('NOTE: landmask exists, ignore the command. Use Stack.set_landmask(None) to allow new landmask downloading')
            return

        # generate the same as DEM grid
        landmask_filename = os.path.join(self.basedir, 'landmask.nc')

        # geographical grid for interferogram area only
        dem = self.get_dem()
        llmin = dem.lon.min().item()
        llmax = dem.lon.max().item()
        ltmin = dem.lat.min().item()
        ltmax = dem.lat.max().item()
        region = f'{llmin}/{llmax}/{ltmin}/{ltmax}'
        if debug:
            print('region', region)

        with tqdm(desc='Landmask Downloading', total=1) as pbar:
            landmask = pygmt.grdlandmask(resolution='f', region=region, spacing=product, maskvalues='NaN/1')
            if os.path.exists(landmask_filename):
                os.remove(landmask_filename)
            landmask.to_netcdf(landmask_filename)
            pbar.update(1)

        self.landmask_filename = landmask_filename

    def plot_landmask(self, landmask='auto', caption='Land Mask', cmap='binary', aspect=None, **kwargs):
        import matplotlib.pyplot as plt

        if isinstance(landmask, str) and landmask == 'auto':
            landmask = self.get_landmask()

        plt.figure()
        landmask.plot.imshow(vmin=0, cmap=cmap)
        self.plot_AOI(**kwargs)
        self.plot_POI(**kwargs)
        if aspect is not None:
            plt.gca().set_aspect(aspect)
        plt.title(caption)



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_geocode import Stack_geocode
from .tqdm_dask import tqdm_dask

class Stack_incidence(Stack_geocode):

    def los_projection(self, data):
        """
        Calculate LOS projection for vector defined by its dx, dy, dz components.

        Parameters
        ----------
        data : xarray dataset
            The input data containing the displacement components dx, dy, dz.

        Returns
        -------
        float, numpy.ndarray, pandas.DataFrame
            The LOS projection. Type of return depends on the input type.

        Examples
        -------
        Calculate tidal LOS projection measured in meter [m]:
        los_projection_mm = stack.los_projection(tidal)
        # Expected input
        # xarray.Dataset
        # Dimensions:
        # date: 31 y: 1 x: 1
        # Data variables:
        # dx (date, y, x) float64 -0.06692 -0.03357 ... 0.005664
        # dy (date, y, x) float64 -0.004765 0.01228 ... -0.04304
        # dz (date, y, x) float64 0.0162 -0.0999 ... 0.005759
        # ...        
        # Expected output:
        # xarray.DataArray date: 31 y: 1 x: 1
        # array([ 0.05532877, -0.05658128, -0.11400223, -0.06658935, -0.0071757 ,
        #    -0.02071992, -0.07211125, -0.12153598, -0.09518547, -0.10037747,
        #    -0.0914933 , -0.12743347, -0.11006747, -0.0643307 , -0.04372583,
        #    -0.07117568, -0.13215618, -0.10467723, -0.01379629,  0.03088265,
        #     0.02786578, -0.01465195, -0.12157386, -0.11801581, -0.001239  ,
        #     0.11614589,  0.07466661, -0.05334002, -0.10686331, -0.06112201,
        #     0.00554765])
        # ...
    
        Calculate plate velocity LOS projection in millimeter [mm]:
        stack.los_projection([22.67, 13.36, 0])
        # Expected output:
        # NOTE: estimation using central point satellite look vector
        # array([-15.57419278])
        """
        import xarray as xr
        import numpy as np

        sat_look = self.get_satellite_look_vector()

        if isinstance(data, xr.Dataset):
            look = sat_look.interp_like(data, method='linear', assume_sorted=True)
            los = xr.dot(xr.concat([look.look_E, look.look_N, look.look_U], dim='dim'),
                   xr.concat([data.dx, data.dy, data.dz], dim='dim'),
                  dims=['dim'])
            return los.transpose('date',...)
        elif isinstance(data, (list, tuple)):
            print ('NOTE: estimation using central point satellite look vector')
            look = sat_look.isel(y=sat_look.y.size//2, x=sat_look.x.size//2)
            data = np.column_stack(data)
            return np.dot(data, [look.look_E, look.look_N, look.look_U])

#     def los_projection(self, data):
#         """
#         Calculate LOS projection for vector defined by its dx, dy, dz components.
# 
#         Parameters
#         ----------
#         data : list, tuple, numpy.ndarray, pandas.DataFrame
#             The input data containing the displacement components dx, dy, dz.
#         scale : float, optional
#             Scale factor to convert input displacements in meter to output LOS displacement in millimeter.
# 
#         Returns
#         -------
#         float, numpy.ndarray, pandas.DataFrame
#             The LOS projection. Type of return depends on the input type.
# 
#         Note
#         ----
#         The function is not optimized for delayed execution.
# 
#         Examples
#         -------
#         Calculate tidal LOS projection:
#         los_projection_mm = 1000*stack.los_projection(tidal)
#         # Expected input
#         #        lon       lat       dx         dy          dz
#         # date						
#         # 2022-06-16  13400.758  47401.431  -66.917528  -4.765059  16.200381
#         # ...        
#         # Expected output:
#         #        lon       lat       dx         dy          dz        los
#         # date						
#         # 2022-06-16  13400.758  47401.431  -66.917528  -4.765059  16.200381  55.340305
#         # ...
# 
#         Using list or tuple as input:
#         los_projection_mm = 1000*stack.los_projection([tidal.dx, tidal.dy, tidal.dz], lon, lat)
#         # Expected output:
#         # [55.34030452, -56.55791618, ...]
# 
#         Using numpy.ndarray as input:
#         los_projection_mm = 1000*stack.los_projection(np.column_stack([tidal.dx, tidal.dy, tidal.dz]))
#         # Expected output (with central point satellite look vector estimation):
#         # [54.72536278, -57.87347137, ...]
# 
#         Note: When lat and lon are not provided, the function will estimate using a central point satellite look vector.
# 
#         """
#         import xarray as xr
#         import pandas as pd
#         import numpy as np
# 
#         sat_look = self.get_sat_look()
# 
#         if isinstance(data, (list, tuple)):
#             data = np.column_stack(data)
# 
#         if isinstance(data, np.ndarray):
#             #if y is not None and x is not None:
#             #    look = sat_look.sel(y=y, x=x, method='nearest')
#             #else:
#             print ('NOTE: estimation using central point satellite look vector')
#             look = sat_look.isel(y=sat_look.y.size//2, x=sat_look.x.size//2)
#             # only for input scalars
#             #return data[0] * sat_look.look_E.values + data[1] * sat_look.look_N.values + data[2] * sat_look.look_U.values
#             return np.dot(data, [look.look_E, look.look_N, look.look_U])
#         elif isinstance(data, pd.DataFrame):
#             # TODO: allow to process multiple coordinates
#             if 'y' in data.columns and 'x' in data.columns:
#                 x = data.loc[data.index[0], 'x']
#                 y = data.loc[data.index[0], 'y']
#                 look = sat_look.sel(y=y, x=x, method='nearest')
#             else:
#                 print ('NOTE: estimation using central point satellite look vector')
#                 look = sat_look.isel(y=sat_look.y.size//2, x=sat_look.x.size//2)
#         elif isinstance(data, xr.Dataset):
#             # TODO: allow to process multiple coordinates
#             if 'y' in data and 'x' in data:
#                 x = data.x.values
#                 y = data.y.values
#                 #print ('y, x', y, x)
#                 look = sat_look.sel(y=y, x=x, method='nearest')
#             else:
#                 print ('NOTE: estimation using central point satellite look vector')
#                 look = sat_look.isel(y=sat_look.y.size//2, x=sat_look.x.size//2)
#         los = np.dot(np.column_stack([data.dx, data.dy, data.dz]), [look.look_E, look.look_N, look.look_U])
#         return los

#     def los_projection(self, data, scale=1):
#         """
#         Calculate LOS projection for vector defined by its dx, dy, dz components.
# 
#         Parameters
#         ----------
#         data : list, tuple, numpy.ndarray, pandas.DataFrame
#             The input data containing the displacement components dx, dy, dz.
#         scale : float, optional
#             Scale factor to convert input displacements in meter to output LOS displacement in millimeter.
# 
#         Returns
#         -------
#         float, numpy.ndarray, pandas.DataFrame
#             The LOS projection. Type of return depends on the input type.
# 
#         Note
#         ----
#         The function is not optimized for delayed execution.
# 
#         Examples
#         -------
#         Calculate tidal LOS projection:
#         los_projection_mm = 1000*stack.los_projection(tidal)
#         # Expected input
#         #        lon       lat       dx         dy          dz
#         # date						
#         # 2022-06-16  13400.758  47401.431  -66.917528  -4.765059  16.200381
#         # ...        
#         # Expected output:
#         #        lon       lat       dx         dy          dz        los
#         # date						
#         # 2022-06-16  13400.758  47401.431  -66.917528  -4.765059  16.200381  55.340305
#         # ...
# 
#         Using list or tuple as input:
#         los_projection_mm = 1000*stack.los_projection([tidal.dx, tidal.dy, tidal.dz], lon, lat)
#         # Expected output:
#         # [55.34030452, -56.55791618, ...]
# 
#         Using numpy.ndarray as input:
#         los_projection_mm = 1000*stack.los_projection(np.column_stack([tidal.dx, tidal.dy, tidal.dz]))
#         # Expected output (with central point satellite look vector estimation):
#         # [54.72536278, -57.87347137, ...]
# 
#         Note: When lat and lon are not provided, the function will estimate using a central point satellite look vector.
# 
#         """
#         import pandas as pd
#         import numpy as np
# 
#         sat_look = self.get_sat_look()
# 
#         if isinstance(data, (list, tuple)):
#             data = np.column_stack(data)
# 
#         if isinstance(data, np.ndarray):
#             #if y is not None and x is not None:
#             #    look = sat_look.sel(y=y, x=x, method='nearest')
#             #else:
#             print ('NOTE: estimation using central point satellite look vector')
#             look = sat_look.isel(y=sat_look.y.size//2, x=sat_look.x.size//2)
#             # only for input scalars
#             #return data[0] * sat_look.look_E.values + data[1] * sat_look.look_N.values + data[2] * sat_look.look_U.values
#             return np.dot(data, [look.look_E, look.look_N, look.look_U])
# 
#         #elif isinstance(data, pd.DataFrame):
#         # TODO: allow to process multiple coordinates
#         if 'y' in data.columns and 'x' in data.columns:
#             x = data.loc[data.index[0], 'x']
#             y = data.loc[data.index[0], 'y']
#             look = sat_look.sel(y=y, x=x, method='nearest')
#         else:
#             print ('NOTE: estimation using central point satellite look vector')
#             look = sat_look.isel(y=sat_look.y.size//2, x=sat_look.x.size//2)
#         los = np.dot(np.column_stack([data.dx, data.dy, data.dz]), [look.look_E, look.look_N, look.look_U])
#         return data.assign(los=scale*los)

    def get_satellite_look_vector(self):
        """
        Return satellite look vectors in geographic coordinates as Xarray Dataset.

        Returns
        -------
        xarray.Dataset
            The satellite look vectors in geographic coordinates.

        Examples
        --------
        Get satellite look vectors:
        sat_look_ll = stack.get_satellite_look_vector()

        Notes
        -----
        This function returns the satellite look vectors in geographic coordinates as Xarray Dataset. The satellite look vectors
        should be computed and saved prior to calling this function using the `sat_look` method.
        """
        return self.open_cube('sat_look')

    #gmt grdmath unwrap_mask.grd $wavel MUL -79.58 MUL = los.grd
    def los_displacement_mm(self, data):
        """
        Compute line-of-sight (LOS) displacement in millimeters.

        Parameters
        ----------
        data : xarray.DataArray or constant, list, tuple, Numpy array, Pandas Series
            Unwrapped phase grid(s) in radar or geographic coordinates.

        Returns
        -------
        xarray.DataArray
            Line-of-sight (LOS) displacement grid(s) in millimeters.

        Examples
        --------
        Calculate LOS displacement for unwrapped phase grids in radar coordinates:
        unwraps_ra = stack.open_grids(pairs, 'unwrap')
        los_disp_ra = stack.los_displacement_mm(unwraps_ra)
        # or the same code in one line
        los_disp_ra = stack.open_grids(pairs, 'unwrap', func=stack.los_displacement_mm)
        # Note: here "func" argument for open_grids() function reduces the code to a single command.

        Calculate LOS displacement for detrended unwrapped phase grids in geographic coordinates:
        detrend_ll = stack.open_grids(pairs, 'detrend', geocode=True)
        los_disp_ll = stack.los_displacement_mm(detrend_ll)
        # or the same code in one line
        los_disp_ll = stack.open_grids(pairs, 'detrend', geocode=True, func=stack.los_displacement_mm)
        # Note: here "func" argument for open_grids() function reduces the code to a single command.
        """
        import xarray as xr
        import numpy as np

        # constant is negative to make LOS = -1 * range change
        # constant is (1000 mm) / (4 * pi)
        scale = -79.58 * self.PRM().get('radar_wavelength')

        if isinstance(data, (list, tuple)):
            return scale*np.asarray(data)
        elif isinstance(data, (xr.DataArray)):
            return (scale*data).rename('los')
        else:
            return scale*data

    def incidence_angle(self):
        """
        Compute the incidence angle grid in geographic coordinates.

        Returns
        -------
        xarray.DataArray
            The incidence angle grid in geographic coordinates.

        Examples
        --------
        Compute the incidence angle grid:
        inc_angle_ll = stack.incidence_angle()

        Notes
        -----
        This function computes the incidence angle grid in geographic coordinates based on the satellite look vectors.
        The satellite look vectors should be computed and saved prior to calling this function using the `sat_look` method.
        The incidence angle is calculated using the formula:
        incidence_angle = arctan2(sqrt(look_E**2 + look_N**2), look_U)
        """
        import xarray as xr
        import numpy as np

        sat_look = self.get_satellite_look_vector()
        incidence_ll = np.arctan2(np.sqrt(sat_look.look_E**2 + sat_look.look_N**2), sat_look.look_U).rename('incidence_angle')
        return incidence_ll

    def plot_incidence_angle(self, incidence_angle='auto', caption='Incidence Angle in Radar Coordinates, [rad]', cmap='gray', aspect=None, **kwargs):
        import matplotlib.pyplot as plt

        plt.figure()
        if isinstance(incidence_angle, str) and incidence_angle == 'auto':
            self.incidence_angle().plot.imshow(cmap=cmap)
        else:
            incidence_angle.plot.imshow(cmap=cmap)
        self.plot_AOI(**kwargs)
        self.plot_POI(**kwargs)
        if aspect is not None:
            plt.gca().set_aspect(aspect)
        plt.xlabel('Range')
        plt.ylabel('Azimuth')
        plt.title(caption)

    def vertical_displacement_mm(self, unwrap):
        """
        Compute vertical displacement in millimeters in radar coordinates.

        Parameters
        ----------
        unwrap : xarray.DataArray or xarray.Dataset
            Unwrapped phase grid(s) in radar coordinates.

        Returns
        -------
        xarray.DataArray
            Vertical displacement grid(s) in millimeters.

        Examples
        --------
        ...
        """
        import numpy as np

        assert self.is_ra(unwrap), 'ERROR: unwrapped phase needs to be defined in radar coordinates'

        los_disp = self.los_displacement_mm(unwrap)
        incidence = self.incidence_angle().reindex_like(los_disp, method='nearest')
        return los_disp/np.cos(incidence)

    def eastwest_displacement_mm(self, unwrap):
        """
        Compute East-West displacement in millimeters.

        Parameters
        ----------
        unwraps : xarray.DataArray or xarray.Dataset
            Unwrapped phase grid(s) in geographic coordinates.

        Returns
        -------
        xarray.DataArray or xarray.Dataset
            East-West displacement grid(s) in millimeters.

        Examples
        --------
        Calculate East-West displacement for unwrapped phase grids in geographic coordinates:
        unwraps_ll = stack.open_grids(pairs, 'unwrap', geocode=True)
        ew_disp_mm = stack.eastwest_displacement_mm(unwraps_ll)

        Calculate East-West displacement for detrended unwrapped phase grids in geographic coordinates:
        ew_disp_mm = stack.open_grids(pairs, 'detrend', geocode=True, func=stack.eastwest_displacement_mm)
        # Note: here "func" argument for open_grids() function reduces the code to a single command.
        """
        import numpy as np
    
        # this displacement is not symmetrical for the orbits due to scene geometries
        orbit = self.df.orbit.unique()[0]
        sign = 1 if orbit == 'D' else -1
        los_disp = self.los_displacement_mm(unwrap)
        incidence_ll = self.incidence_angle().reindex_like(los_disp, method='nearest')
        return sign * los_disp/np.sin(incidence_ll)

    def compute_satellite_look_vector(self, interactive=False):
        #import dask
        import xarray as xr
        import numpy as np

        # ..., look_E, look_N, look_U
        satlook_map = {0: 'look_E', 1: 'look_N', 2: 'look_U'}

        def SAT_look(z, lat, lon):
            coords = np.column_stack([lon.ravel(), lat.ravel(), z.ravel()])
            # look_E look_N look_U
            look = self.PRM().SAT_look(coords, binary=True)\
                                     .astype(np.float32)\
                                     .reshape(z.shape[0], z.shape[1], 6)[...,3:]
            return look

        # reference grid
        trans_inv = self.get_trans_inv()[['lt', 'll', 'ele']]

        # xarray wrapper for the valid area only
        enu = xr.apply_ufunc(
            SAT_look,
            trans_inv.ele,
            trans_inv.lt,
            trans_inv.ll,
            dask='parallelized',
            vectorize=False,
            output_dtypes=[np.float32],
            output_core_dims=[['enu']],
            dask_gufunc_kwargs={'output_sizes': {'enu': 3}}
        )

        # transform to separate variables
        keys_vars = {val: enu[...,key] for (key, val) in satlook_map.items()}
        sat_look = xr.Dataset(keys_vars)

        if interactive:
            return sat_look

        return self.save_cube(sat_look, 'sat_look', 'Satellite Look Vector Computing')



# ----------------------------------------------------------------------------
# PyGMTSAR
#.
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
#.
# Copyright (c) 2023, Alexey Pechnikov
#.
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .datagrid import datagrid
from pygmtsar import tqdm_dask

class IO(datagrid):

    # processing directory
    basedir = '.'

    def _glob_re(self, pathname):
        import os
        import re
        filenames = filter(re.compile(pathname).match, os.listdir(self.basedir))
        return sorted([os.path.join(self.basedir, filename) for filename in filenames])

    def dump(self, to_path=None):
        """
        Dump Stack object state to a pickle file (stack.pickle in the processing directory by default).

        Parameters
        ----------
        to_path : str, optional
            Path to the output dump file. If not provided, the default dump file in the processing directory is used.

        Returns
        -------
        None

        Examples
        --------
        Dump the current state to the default dump file in the processing directory:
        stack.dump()

        Notes
        -----
        This method serializes the state of the Stack object and saves it to a pickle file. The pickle file can be used to
        restore the Stack object with its processed data and configuration. By default, the dump file is named "stack.pickle"
        and is saved in the processing directory. An alternative file path can be provided using the `to_path` parameter.
        """
        import pickle
        import os

        if to_path is None:
            stack_pickle = os.path.join(self.basedir, 'stack.pickle')
        else:
            if os.path.isdir(to_path):
                stack_pickle = os.path.join(to_path, 'stack.pickle')
            else:
                stack_pickle = to_path
    
        print (f'NOTE: save state to file {stack_pickle}')
        pickle.dump(self, open(stack_pickle, 'wb'))

        return

    @staticmethod
    def restore(from_path):
        """
        Restore Stack object state from a pickle file (stack.pickle in the processing directory by default).

        Parameters
        ----------
        from_path : str
            Path to the input dump file.

        Returns
        -------
        Stack
            The restored Stack object.

        Examples
        --------
        Restore the current state from the default dump file in the processing directory:
        Stack.restore()

        Notes
        -----
        This static method restores the state of an Stack object from a pickle file. The pickle file should contain the
        serialized state of the Stack object, including its processed data and configuration. By default, the method assumes
        the input file is named "stack.pickle" and is located in the processing directory. An alternative file path can be
        provided using the `from_path` parameter. The method returns the restored Stack object.
        """
        import pickle
        import os

        if os.path.isdir(from_path):
            stack_pickle = os.path.join(from_path, 'stack.pickle')
        else:
            stack_pickle = from_path

        print (f'NOTE: load state from file {stack_pickle}')
        return pickle.load(open(stack_pickle, 'rb'))

    def backup(self, backup_dir, copy=False, debug=False):
        """
        Backup framed Stack scenes, orbits, DEM, and landmask files to build a minimal reproducible dataset.

        Parameters
        ----------
        backup_dir : str
            The backup directory where the files will be copied.
        copy : bool, optional
            Flag indicating whether to make a copy of scene and orbit files and remove the originals in the work directory.
            If False, the files will be moved to the backup directory. Default is False.
        debug : bool, optional
            Flag indicating whether to print debug information. Default is False.

        Returns
        -------
        None

        Examples
        --------
        Backup the files to the specified directory:
        stack.backup('backup')

        Open the backup for the reproducible run by defining it as a new data directory:
        stack = Stack('backup', 'backup/DEM_WGS84.nc', 'raw')

        Notes
        -----
        This method backs up the framed Stack scenes, orbits, DEM, and landmask files to a specified backup directory.
        It provides a way to create a minimal reproducible dataset by preserving the necessary files for processing.
        The method creates the backup directory if it does not exist. By default, the method moves the scene and orbit files
        to the backup directory, effectively removing them from the work directory. The DEM and landmask files are always
        copied to the backup directory. If the `copy` parameter is set to True, the scene and orbit files will be copied
        instead of moved. Use caution when setting `copy` to True as it can result in duplicated files and consume
        additional storage space. The method also updates the Stack object's dataframe to mark the removed files as empty.
        """
        import os
        import shutil

        os.makedirs(backup_dir, exist_ok=True)

        # save the geometry
        filename = os.path.join(backup_dir, 'stack.geojson')
        self.df[['datetime', 'orbit','mission','polarization', 'subswath','geometry']].to_file(filename, driver='GeoJSON')

        # this optional file is dumped state, copy it if exists
        # auto-generated file can't be a symlink but user-defined symlink target should be copied
        filename = os.path.join(self.basedir, 'stack.pickle')
        if os.path.exists(filename):
            if debug:
                print ('DEBUG: copy', filename, backup_dir)
            shutil.copy2(filename, backup_dir, follow_symlinks=True)

        # these files required to continue the processing, do not remove and copy only
        filenames = [self.dem_filename, self.landmask_filename]
        for filename in filenames:
            # DEM and landmask can be not defined
            if filename is None:
                continue
            if debug:
                print ('DEBUG: copy', filename, backup_dir)
            shutil.copy2(filename, backup_dir, follow_symlinks=True)

        # these files are large and are not required to continue the processing
        filenames = []
        for record in self.df.itertuples():
            for filename in [record.datapath, record.metapath, record.orbitpath]:
                filenames += filename if isinstance(filename, list) else [filename]
        for filename in filenames:
            # orbit files can be not defined
            if filename is None:
                continue
            # copy and delete the original later to prevent cross-device links issues
            if debug:
                print ('DEBUG: copy', filename, backup_dir)
            shutil.copy2(filename, backup_dir, follow_symlinks=True)
            if not copy and self.basedir == os.path.dirname(filename):
                # when copy is not needed then delete files in work directory only
                if debug:
                    print ('DEBUG: remove', filename)
                os.remove(filename)

        if not copy:
            # mark as empty all the removed files
            for col in ['datapath','metapath','orbitpath']:
                self.df[col] = None
        return

    def get_filename(self, name, add_subswath=False):
        import os

        if add_subswath:
            subswath = self.get_subswath()
            prefix = f'F{subswath}_'
        else:
            prefix = ''

        filename = os.path.join(self.basedir, f'{prefix}{name}.grd')
        return filename

    def get_filenames(self, pairs, name, add_subswath=False):
        """
        Get the filenames of the data grids. The filenames are determined by the subswath, pairs, and name parameters.

        Parameters
        ----------
        subswath : int or None
            The subswath number. If None, the function will determine it based on the dataset.
        pairs : np.ndarray or pd.DataFrame or None
            An array or DataFrame of pairs. If None, the function will open a single grid or a set of subswath grids.
        name : str
            The name of the grid to be opened.
        add_subswath : bool, optional
            Whether to add subswath to the prefix of the filename. Default is True.

        Returns
        -------
        str or list of str
            The filename or a list of filenames of the grids.
        """
        import pandas as pd
        import numpy as np
        import os

        if add_subswath:
            subswath = self.get_subswath()
            prefix = f'F{subswath}_'
        else:
            prefix = ''

        if isinstance(pairs, pd.DataFrame):
            # convert to standalone DataFrame first
            pairs = self.get_pairs(pairs)[['ref', 'rep']].astype(str).values
        else:
            pairs = np.asarray(pairs)

        filenames = []
        if len(pairs.shape) == 1:
            # read all the grids from files
            for date in sorted(pairs):
                filename = os.path.join(self.basedir, f'{prefix}{name}_{date}.grd'.replace('-',''))
                filenames.append(filename)
        elif len(pairs.shape) == 2:
            # read all the grids from files
            for pair in pairs:
                filename = os.path.join(self.basedir, f'{prefix}{name}_{pair[0]}_{pair[1]}.grd'.replace('-',''))
                filenames.append(filename)
        return filenames

    # 2.5e-07 is Sentinel-1 scale factor
    def open_data(self, dates=None, scale=2.5e-07, debug=False):
        import xarray as xr
        import pandas as pd
        import numpy as np
        import os

        if debug:
            print ('DEBUG: open_data: apply scale:', scale)

        if dates is None:
            dates = self.df.index.values
        #print ('dates', dates)

        filenames = [self.PRM(date).filename[:-4] + '.grd' for date in dates]
        #print ('filenames', filenames)
        ds = xr.open_mfdataset(
            filenames,
            engine=self.netcdf_engine,
            chunks=self.chunksize,
            parallel=True,
            concat_dim='date',
            combine='nested'
        ).assign(date=pd.to_datetime(dates)).rename({'a': 'y', 'r': 'x'})
        if scale is None:
            # there is no complex int16 datatype, so return two variables for real and imag parts
            return ds
        # scale and return as complex values
        ds_scaled = (scale*(ds.re.astype(np.float32) + 1j*ds.im.astype(np.float32))).assign_attrs(ds.attrs).rename('data')
        del ds
        # zero in np.int16 type means NODATA
        return ds_scaled.where(ds_scaled != 0)

#     def open_geotif(self, dates=None, subswath=None, intensity=False, chunksize=None):
#         """
#         tiffs = stack.open_data_geotif(['2022-06-16', '2022-06-28'], intensity=True)
#         """
#         import xarray as xr
#         import rioxarray as rio
#         import pandas as pd
#         import numpy as np
#         # from GMTSAR code
#         DFACT = 2.5e-07
#     
#         if chunksize is None:
#             chunksize = self.chunksize
# 
#         if subswath is None:
#             subswaths = self.get_subswaths()
#         else:
#             subswaths = [subswath]
# 
#         stack = []
#         for swath in self.get_subswaths():
#             if dates is None:
#                 dates = self.df[self.df['subswath']==swath].index.values
#             #print ('dates', dates)
#             tiffs = self.df[(self.df['subswath']==swath)&(self.df.index.isin(dates))].datapath.values
#             tiffs = [rio.open_rasterio(tif, chunks=chunksize)[0] for tif in tiffs]
#             # build stack
#             tiffs = xr.concat(tiffs, dim='date')
#             tiffs['date'] = pd.to_datetime(dates)
#             # 2 and 4 multipliers to have the same values as in SLC
#             if intensity:
#                 stack.append((2*DFACT*np.abs(tiffs))**2)
#             else:
#                 stack.append(2*DFACT*tiffs)
# 
#         return stacks if subswath is None else stacks[0]

    def open_cube(self, name):
        """
        Opens an xarray 2D/3D Dataset or dataArray from a NetCDF file.

        This function takes the name of the model to be opened, reads the NetCDF file, and re-chunks
        the dataset according to the provided chunksize or the default value from the 'stack' object.
        The 'date' dimension is always chunked with a size of 1.

        Parameters
        ----------
        name : str
            The name of the model file to be opened.

        Returns
        -------
        xarray.Dataset
            Xarray Dataset read from the specified NetCDF file.

        """
        import xarray as xr
        import pandas as pd
        import numpy as np
        import os

        filename = self.get_filename(name)
        assert os.path.exists(filename), f'ERROR: The NetCDF file is missed: {filename}'

        # Workaround: open the dataset without chunking
        data = xr.open_dataset(filename, engine=self.netcdf_engine)
        # Determine the proper chunk sizes
        chunks = {dim: 1 if dim in ['pair', 'date'] else self.chunksize for dim in data.dims}
        # Re-chunk the dataset using the chunks dictionary
        data = data.chunk(chunks)

        # attributes are empty when dataarray is prezented as dataset
        # revert dataarray converted to dataset
        data_vars = list(data.data_vars)
        if len(data_vars) == 1 and 'dataarray' in data.attrs:
            assert data.attrs['dataarray'] == data_vars[0]
            data = data[data_vars[0]]

        # convert string dates to dates
        for dim in ['date', 'ref', 'rep']:
            if dim in data.dims:
                data[dim] = pd.to_datetime(data[dim])

        # restore missed coordinates
        for dim in ['y', 'x', 'lat', 'lon']:
            if dim not in data.coords \
                and f'start_{dim}' in data.attrs \
                and f'step_{dim}' in data.attrs \
                and f'stop_{dim}' in data.attrs \
                and f'size_{dim}' in data.attrs:
                start = data.attrs[f'start_{dim}']
                step  = data.attrs[f'step_{dim}']
                stop  = data.attrs[f'stop_{dim}']
                size  = data.attrs[f'size_{dim}']
                coords = np.arange(start, stop+step/2, step)
                assert coords.size == size, 'Restored dimension has invalid size'
                #print (dim, start, step, stop, coords)
                data = data.assign_coords({dim: coords})
                # remove the system attributes
                del data.attrs[f'start_{dim}']
                del data.attrs[f'step_{dim}']
                del data.attrs[f'stop_{dim}']
                del data.attrs[f'size_{dim}']

        return data

    def sync_cube(self, data, name=None, caption='Syncing NetCDF 2D/3D Dataset'):
        import xarray as xr
        if name is None and isinstance(data, xr.DataArray):
            assert data.name is not None, 'Define data name or use "name" argument for the NetCDF filename'
            name = data.name
        elif name is None:
            raise ValueError('Specify name for the output NetCDF file')
        self.save_cube(data, name, caption)
        return self.open_cube(name)

    def save_cube(self, data, name=None, caption='Saving NetCDF 2D/3D Dataset'):
        """
        Save a lazy and not lazy 2D/3D xarray Dataset or DataArray to a NetCDF file.

        The 'date' or 'pair' dimension is always chunked with a size of 1.

        Parameters
        ----------
        data : xarray.Dataset or xarray.DataArray
            The model to be saved.
        name : str
            The text name for the output NetCDF file.
        caption: str
            The text caption for the saving progress bar.

        Returns
        -------
        None

        Examples
        -------
        stack.save_cube(intf90m, 'intf90m')                              # save lazy 3d dataset
        stack.save_cube(intf90m.phase, 'intf90m')                        # save lazy 3d dataarray
        stack.save_cube(intf90m.isel(pair=0), 'intf90m')                 # save lazy 2d dataset
        stack.save_cube(intf90m.isel(pair=0).phase, 'intf90m')           # save lazy 2d dataarray
        stack.save_cube(intf90m.compute(), 'intf90m')                    # save 3d dataset     
        stack.save_cube(intf90m.phase.compute(), 'intf90m')              # save 3d dataarray
        stack.save_cube(intf90m.isel(pair=0).compute(), 'intf90m')       # save 2d dataset
        stack.save_cube(intf90m.isel(pair=0).phase.compute(), 'intf90m') # save 2d dataarray
        """
        import xarray as xr
        import dask
        import os
        import warnings
        # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
        warnings.filterwarnings('ignore')
        warnings.filterwarnings('ignore', module='dask')
        warnings.filterwarnings('ignore', module='dask.core')
        import logging
        # prevent warnings "RuntimeWarning: All-NaN slice encountered"
        logging.getLogger('distributed.nanny').setLevel(logging.ERROR)
        # disable "distributed.utils_perf - WARNING - full garbage collections ..."
        from dask.distributed import utils_perf
        utils_perf.disable_gc_diagnosis()

        if name is None and isinstance(data, xr.DataArray):
            assert data.name is not None, 'Define data name or use "name" argument for the NetCDF filename'
            name = data.name
        elif name is None:
            raise ValueError('Specify name for the output NetCDF file')

        for dim in ['y', 'x', 'lat', 'lon']:
            if dim in data.dims:
                # use attributes to hold grid spacing to prevent xarray Dask saving issues
                data = data.drop(dim, dim=None).assign_attrs({
                    f'start_{dim}': data[dim].values[0],
                    f'step_{dim}': data[dim].diff(dim).values[0],
                    f'stop_{dim}': data[dim].values[-1],
                    f'size_{dim}': data[dim].size
                })

        if isinstance(data, xr.DataArray):
            if data.name is None:
                data = data.rename(name)
            data = data.to_dataset().assign_attrs({'dataarray': data.name})

        is_dask = isinstance(data[list(data.data_vars)[0]].data, dask.array.Array)
        encoding = {varname: self._compression(data[varname].shape) for varname in data.data_vars}
        #print ('is_dask', is_dask, 'encoding', encoding)

        # save to NetCDF file
        filename = self.get_filename(name)
        if os.path.exists(filename):
            os.remove(filename)
        delayed = data.to_netcdf(filename,
                                 engine=self.netcdf_engine,
                                 encoding=encoding,
                                 compute=not is_dask)
        if is_dask:
            tqdm_dask(result := dask.persist(delayed), desc=caption)
            # cleanup - sometimes writing NetCDF handlers are not closed immediately and block reading access
            del delayed, result
            import gc; gc.collect()

    def delete_cube(self, name):
        import os

        filename = self.get_filename(name)
        #print ('filename', filename)
        if os.path.exists(filename):
            os.remove(filename)

    def sync_stack(self, data, name=None, caption='Saving 2D Stack', queue=None, timeout=300):
        import xarray as xr
        if name is None and isinstance(data, xr.DataArray):
            assert data.name is not None, 'Define data name or use "name" argument for the NetCDF filenames'
            name = data.name
        elif name is None:
            raise ValueError('Specify name for the output NetCDF files')
        self.delete_stack(name)
        self.save_stack(data, name, caption, queue, timeout)
        return self.open_stack(name)

    def open_stack(self, name, stack=None):
        """
        Examples:
        stack.open_stack('data')
        stack.open_stack('data', ['2018-03-23'])
        stack.open_stack('data', ['2018-03-23', '2018-03-11'])
        stack.open_stack('phase15m')
        stack.open_stack('intf90m',[['2018-02-21','2018-03-11']])
        stack.open_stack('intf90m', stack.get_pairs([['2018-02-21','2018-03-11']]))
        """
        import xarray as xr
        import pandas as pd
        import numpy as np
        import glob

        if stack is None:
            # look for all stack files
            #filenames = self.get_filenames(['*'], name)[0]
            #filenames = self.get_filename(f'{name}_????????_????????')
            # like data_20180323.grd or intf60m_20230114_20230219.grd
            filenames = self._glob_re(name + '_[0-9]{8}(_[0-9]{8})*.grd')
        elif isinstance(stack, (list, tuple, np.ndarray)) and len(np.asarray(stack).shape) == 1:
            # dates
            filenames = self.get_filenames(np.asarray(stack), name)
        else:
            # pairs
            filenames = self.get_filenames(stack, name)
        #print ('filenames', filenames)

        data = xr.open_mfdataset(
            filenames,
            engine=self.netcdf_engine,
            chunks=self.chunksize,
            parallel=True,
            concat_dim='stackvar',
            combine='nested'
        )

        # revert dataarray converted to dataset
        data_vars = list(data.data_vars)
        if len(data_vars) == 1 and 'dataarray' in data.attrs:
            assert data.attrs['dataarray'] == data_vars[0]
            data = data[data_vars[0]]

        # attributes are empty when dataarray is prezented as dataset, convert it first
        # restore missed coordinates
        for dim in ['y', 'x', 'lat', 'lon']:
            if dim not in data.coords \
                and f'start_{dim}' in data.attrs \
                and f'step_{dim}' in data.attrs \
                and f'stop_{dim}' in data.attrs \
                and f'size_{dim}' in data.attrs:
                start = data.attrs[f'start_{dim}']
                step  = data.attrs[f'step_{dim}']
                stop  = data.attrs[f'stop_{dim}']
                size  = data.attrs[f'size_{dim}']
                coords = np.arange(start, stop+step/2, step)
                assert coords.size == size, 'Restored dimension has invalid size'
                #print (dim, start, step, stop, coords)
                data = data.assign_coords({dim: coords})
                # remove the system attributes
                del data.attrs[f'start_{dim}']
                del data.attrs[f'step_{dim}']
                del data.attrs[f'stop_{dim}']
                del data.attrs[f'size_{dim}']

        for dim in ['pair', 'date']:
            if dim in data.coords:
                if data[dim].shape == ():
                    data = data.rename({'stackvar': dim})
                else:
                    data = data.swap_dims({'stackvar': dim})

        # convert string (or already timestamp) dates to dates
        for dim in ['date', 'ref', 'rep']:
            if dim in data.dims:
                if not data[dim].shape == ():
                    data[dim] = pd.to_datetime(data[dim])
                else:
                    data[dim].values = pd.to_datetime(data['date'].values)

        return data

#     # simple sequential realization is not suitable for a large stack
#     def save_stack(self, data, name, caption='Saving 2D grids stack'):
#         import xarray as xr
#         import dask
#         import os
#         import warnings
#         # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
#         warnings.filterwarnings('ignore')
#         warnings.filterwarnings('ignore', module='dask')
#         warnings.filterwarnings('ignore', module='dask.core')
# 
#         if isinstance(data, xr.Dataset):
#             stackvar = data[list(data.data_vars)[0]].dims[0]
#             is_dask = isinstance(data[list(data.data_vars)[0]].data, dask.array.Array)
#         elif isinstance(data, xr.DataArray):
#             stackvar = data.dims[0]
#             is_dask = isinstance(data.data, dask.array.Array)
#         else:
#             raise Exception('Argument grid is not xr.Dataset or xr.DataArray object')
#         #print ('is_dask', is_dask, 'stackvar', stackvar)
#         stacksize = data[stackvar].size
# 
#         for dim in ['y', 'x', 'lat', 'lon']:
#             if dim in data.dims:
#                 # use attributes to hold grid spacing to prevent xarray Dask saving issues
#                 data = data.drop(dim, dim=None).assign_attrs({
#                     f'start_{dim}': data[dim].values[0],
#                     f'step_{dim}':  data[dim].diff(dim).values[0],
#                     f'stop_{dim}':  data[dim].values[-1],
#                     f'size_{dim}':  data[dim].size
#                 })
# 
#         delayeds = []
#         digits = len(str(stacksize))
#         for ind in range(stacksize):
#             data_slice = data.isel({stackvar: ind})
#             if stackvar == 'date':
#                 stackval = str(data_slice[stackvar].dt.date.values)
#             else:
#                 stackval = data_slice[stackvar].item().split(' ')
#             #print ('stackval', stackval)
#             # save to NetCDF file
#             filename = self.get_filenames([stackval], name)[0]
#             #print ('filename', filename)
#             if os.path.exists(filename):
#                 os.remove(filename)
#             if isinstance(data, xr.Dataset):
#                 encoding = {varname: self._compression(data_slice[varname].shape) for varname in data.data_vars}
#             elif isinstance(data, xr.DataArray):
#                 encoding = {data.name: self._compression(data_slice.shape)}
#             else:
#                 raise Exception('Argument grid is not xr.Dataset or xr.DataArray object')
# 
#             delayed = data_slice.to_netcdf(filename,
#                                   encoding=encoding,
#                                   engine=self.netcdf_engine,
#                                   compute=not is_dask)
#             if is_dask:
#                 delayeds.append(delayed)
#                 del delayed
#             del data_slice
# 
#         if is_dask:
#             tqdm_dask(result := dask.persist(*delayeds), desc=caption)
#             # cleanup - sometimes writing NetCDF handlers are not closed immediately and block reading access
#             del delayeds, result
#             import gc; gc.collect()

    # use save_mfdataset
    def save_stack(self, data, name, caption='Saving 2D Stack', queue=None, timeout=300):
        import numpy as np
        import xarray as xr
        import dask
        import os
        from dask.distributed import get_client
        import warnings
        # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
        warnings.filterwarnings('ignore')
        warnings.filterwarnings('ignore', module='dask')
        warnings.filterwarnings('ignore', module='dask.core')
        # Filter out Dask "Restarting worker" warnings
        warnings.filterwarnings("ignore", module="distributed.nanny")
        import logging
        # Suppress Dask "Restarting worker" warnings
        logging.getLogger('distributed.nanny').setLevel(logging.ERROR)
        # disable "distributed.utils_perf - WARNING - full garbage collections ..."
        from dask.distributed import utils_perf
        utils_perf.disable_gc_diagnosis()

        # Dask cluster client
        client = get_client()
        
        if isinstance(data, xr.Dataset):
            stackvar = data[list(data.data_vars)[0]].dims[0]
            is_dask = isinstance(data[list(data.data_vars)[0]].data, dask.array.Array)
        elif isinstance(data, xr.DataArray):
            stackvar = data.dims[0]
            is_dask = isinstance(data.data, dask.array.Array)
        else:
            raise Exception('Argument grid is not xr.Dataset or xr.DataArray object')
        #print ('is_dask', is_dask, 'stackvar', stackvar)
        stacksize = data[stackvar].size

        if queue is None:
            queue = self.netcdf_queue
        if queue is None:
            # process all the stack items in a single operation
            queue = stacksize

        for dim in ['y', 'x', 'lat', 'lon']:
            if dim in data.dims:
                # use attributes to hold grid spacing to prevent xarray Dask saving issues
                data = data.drop(dim, dim=None).assign_attrs({
                    f'start_{dim}': data[dim].values[0],
                    f'step_{dim}':  data[dim].diff(dim).values[0],
                    f'stop_{dim}':  data[dim].values[-1],
                    f'size_{dim}':  data[dim].size
                })

        if isinstance(data, xr.DataArray):
            data = data.to_dataset().assign_attrs({'dataarray': data.name})
        encoding = {varname: self._compression(data[varname].shape[1:]) for varname in data.data_vars}
        #print ('encoding', encoding)

        # Applying iterative processing to prevent Dask scheduler deadlocks.
        counter = 0
        digits = len(str(stacksize))
        # Splitting all the pairs into chunks, each containing approximately queue pairs.
        n_chunks = stacksize // queue if stacksize >= queue else 1
        for chunk in np.array_split(range(stacksize), n_chunks):
            dss = [data.isel({stackvar: ind}) for ind in chunk]
            if stackvar == 'date':
                stackvals = [ds[stackvar].dt.date.values for ds in dss]
            else:
                stackvals = [ds[stackvar].item().split(' ') for ds in dss]
            # save to NetCDF file
            filenames = self.get_filenames(stackvals, name)
            #[os.remove(filename) for filename in filenames if os.path.exists(filename)]
            delayeds = xr.save_mfdataset(dss,
                                         filenames,
                                         encoding=encoding,
                                         engine=self.netcdf_engine,
                                         compute=not is_dask)
            # process lazy chunk
            if is_dask:
                if n_chunks > 1:
                    chunk_caption = f'{caption}: {(counter+1):0{digits}}...{(counter+len(chunk)):0{digits}} from {stacksize}'
                else:
                    chunk_caption = caption
                tqdm_dask(result := dask.persist(delayeds), desc=chunk_caption)
                del delayeds, result
                # cleanup - sometimes writing NetCDF handlers are not closed immediately and block reading access
                import gc; gc.collect()
                # cleanup - release all workers memory, call garbage collector before to prevent heartbeat errors
                if timeout is not None:
                    client.restart(timeout=timeout, wait_for_workers=True)
#                 # more granular control
#                 n_workers = len(client.nthreads())
#                 client.restart(wait_for_workers=False)
#                 client.wait_for_workers(n_workers, timeout=timeout)
            # update chunks counter
            counter += len(chunk)

#     # alternative realization
#     def save_stack(self, data, name, caption='Saving 2D Stack', queue=50):
#         import numpy as np
#         import xarray as xr
#         import dask
#         import os
#         from dask.distributed import get_client
#         import warnings
#         # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
#         warnings.filterwarnings('ignore')
#         warnings.filterwarnings('ignore', module='dask')
#         warnings.filterwarnings('ignore', module='dask.core')
#         # Filter out Dask "Restarting worker" warnings
#         warnings.filterwarnings("ignore", category=UserWarning, module="distributed.nanny")
#         import logging
#         # Suppress Dask "Restarting worker" warnings
#         logging.getLogger('distributed.nanny').setLevel(logging.ERROR)
#         # disable "distributed.utils_perf - WARNING - full garbage collections ..."
#         from dask.distributed import utils_perf
#         utils_perf.disable_gc_diagnosis()
# 
#         if isinstance(data, xr.Dataset):
#             stackvar = data[list(data.data_vars)[0]].dims[0]
#             is_dask = isinstance(data[list(data.data_vars)[0]].data, dask.array.Array)
#         elif isinstance(data, xr.DataArray):
#             stackvar = data.dims[0]
#             is_dask = isinstance(data.data, dask.array.Array)
#         else:
#             raise Exception('Argument grid is not xr.Dataset or xr.DataArray object')
#         #print ('is_dask', is_dask, 'stackvar', stackvar)
#         stacksize = data[stackvar].size
# 
#         if queue is None:
#             # process all the stack items in a single operation
#             queue = stacksize
# 
#         for dim in ['y', 'x', 'lat', 'lon']:
#             if dim in data.dims:
#                 # use attributes to hold grid spacing to prevent xarray Dask saving issues
#                 data = data.drop(dim, dim=None).assign_attrs({
#                     f'start_{dim}': data[dim].values[0],
#                     f'step_{dim}':  data[dim].diff(dim).values[0],
#                     f'stop_{dim}':  data[dim].values[-1],
#                     f'size_{dim}':  data[dim].size
#                 })
# 
#         if isinstance(data, xr.Dataset):
#             encoding = {varname: self._compression(data[varname].shape[1:]) for varname in data.data_vars}
#         elif isinstance(data, xr.DataArray):
#             encoding = {data.name: self._compression(data.shape[1:])}
#         else:
#             raise Exception('Argument data is not 3D xr.Dataset or xr.DataArray object')
#         #print ('encoding', encoding)
# 
#         # Applying iterative processing to prevent Dask scheduler deadlocks.
#         counter = 0
#         digits = len(str(stacksize))
#         # Splitting all the pairs into chunks, each containing approximately queue pairs.
#         n_chunks = stacksize // queue if stacksize >= queue else 1
#         for chunk in np.array_split(range(stacksize), n_chunks):
#             stack = [data.isel({stackvar: ind}) for ind in chunk]
#             assert len(stack) == len(chunk), 'ERROR: incorrect queue size'
#             if stackvar == 'date':
#                 stackvals = [ds[stackvar].dt.date.values for ds in stack]
#             else:
#                 stackvals = [ds[stackvar].item().split(' ') for ds in stack]
#             # NetCDF files
#             filenames = self.get_filenames(stackvals, name)
# 
#             delayeds = []
#             # prepare lazy chunk or process not lazy chunk
#             for data_slice, filename in zip(stack, filenames):
#                 #print ('filename', filename)
#                 if os.path.exists(filename):
#                     os.remove(filename)
#                 delayed = data_slice.to_netcdf(filename,
#                                       encoding=encoding,
#                                       engine=self.netcdf_engine,
#                                       compute=not is_dask)
#                 delayeds.append(delayed)
#                 del delayed, data_slice, filename
#             # process lazy chunk
#             if is_dask:
#                 chunk_caption = f'{caption}: {(counter+1):0{digits}}...{(counter+len(chunk)):0{digits}} from {stacksize}'
#                 tqdm_dask(dask.persist(delayeds), desc=chunk_caption)
#             # update chunks counter
#             counter += len(chunk)
#             del delayeds, stack
#             # cleanup - release all memory
#             get_client().restart()
#         # cleanup - sometimes writing NetCDF handlers are not closed immediately and block reading access
#         import gc; gc.collect()

    def delete_stack(self, name):
        import os

        filenames = self._glob_re(name + '_[0-9]{8}(_[0-9]{8})*.grd')
        #print ('filenames', filenames)
        for filename in filenames:
            if os.path.exists(filename):
                os.remove(filename)



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_unwrap import Stack_unwrap

class Stack_detrend(Stack_unwrap):
# 
#     @staticmethod
#     def regression_linear(data, grid, fit_intercept=True):
#         import numpy as np
#         import xarray as xr
#         import dask
# 
#         # define topography on the same grid and fill NaNs
#         grid = grid.reindex_like(data, method='nearest').fillna(0)
# 
#         # find stack dim
#         stackvar = data.dims[0] if len(data.dims) == 3 else 'stack'
#         #print ('stackvar', stackvar)
#         shape2d = data.shape[1:] if len(data.dims) == 3 else data.shape
#         #print ('shape2d', shape2d)
# 
#         @dask.delayed
#         def regression_block(stackval, fit_intercept):
#             from sklearn.linear_model import LinearRegression
#             from sklearn.pipeline import make_pipeline
#             from sklearn.preprocessing import StandardScaler
# 
#             # use outer variable
#             data_block = (data.sel({stackvar: stackval}) if stackval is not None else data).compute(n_workers=1)
# 
#             grid_values = grid.values.ravel()
#             data_values = data_block.values.ravel()
#             nanmask = np.isnan(grid_values) | np.isnan(data_values)
#         
#             # build prediction model with or without plane removal (fit_intercept)
#             regr = make_pipeline(StandardScaler(), LinearRegression(fit_intercept=fit_intercept))
#             # fit 1D non-NaNs phase and topography and predict on the non-NaNs 2D topography grid
#             data_grid = regr.fit(np.column_stack([grid_values [~nanmask]]),
#                                  np.column_stack([data_values[~nanmask]]))\
#                         .predict(np.column_stack([grid_values])).reshape(data_block.shape)
#             # cleanup
#             del grid_values, data_values, nanmask, data_block
#             return data_grid
#         
#         stack = []
#         for stackval in data[stackvar].values if len(data.dims) == 3 else [None]:
#             #print ('stackval', stackval)
#             block = dask.array.from_delayed(regression_block(stackval, fit_intercept=fit_intercept),
#                                             shape=shape2d, dtype=np.float32)
#             stack.append(block)
#             del block
# 
#         return xr.DataArray(dask.array.stack(stack) if len(data.dims) == 3 else stack[0],
#                             coords=data.coords)\
#                .rename(data.name)

#     @staticmethod
#     def regression_linear(data, variables, weight=None, valid_pixels_threshold=10000, fit_intercept=True):
#         """
#         topo = sbas.get_topo().coarsen({'x': 4}, boundary='trim').mean()
#         yy, xx = xr.broadcast(topo.y, topo.x)
#         strat_sbas = sbas.regression_linear(unwrap_sbas.phase,
#                 [topo,    topo*yy,    topo*xx,    topo*yy*xx,
#                  topo**2, topo**2*yy, topo**2*xx, topo**2*yy*xx,
#                  topo**3, topo**3*yy, topo**3*xx, topo**3*yy*xx,
#                  yy, xx,
#                  yy**2, xx**2, yy*xx,
#                  yy**3, xx**3, yy**2*xx, xx**2*yy], corr_sbas)
#         """
#         import numpy as np
#         import xarray as xr
#         import dask
#         from sklearn.linear_model import LinearRegression
#         from sklearn.pipeline import make_pipeline
#         from sklearn.preprocessing import StandardScaler
# 
#         if weight is not None and len(weight.dims) == 3 and weight.shape != data.shape:
#             raise Exception(f'Argument "weight" 3D shape {weight.shape} should be equal to "data" 3D shape {data.shape}')
#         if weight is not None and len(weight.dims) == 2 and weight.shape != data.shape[1:]:
#             raise Exception(f'Argument "weight" 2D shape {weight.shape} should be equal to "data" 2D shape {data.shape[1:]}')
# 
#         # find stack dim
#         stackvar = data.dims[0] if len(data.dims) == 3 else 'stack'
#         #print ('stackvar', stackvar)
#         shape2d = data.shape[1:] if len(data.dims) == 3 else data.shape
#         #print ('shape2d', shape2d)
#         chunk2d = data.chunks[1:] if len(data.dims) == 3 else data.chunks
#         #print ('chunk2d', chunk2d)
#     
#         if isinstance(variables, (list, tuple)):
#             variables = xr.concat(variables, dim='stack')
#         elif not isinstance(variables, xr.DataArray) or len(variables.dims) not in (2, 3):
#             raise Exception('Argument "variables" should be 2D or 3D Xarray dataarray of list of 2D Xarray dataarrays')
#         elif len(variables.dims) == 2:
#             variables = variables.expand_dims('stack')
#         elif len(variables.dims) == 3 and not variables.dims[0] == 'stack':
#             raise Exception('Argument "variables" 3D Xarray dataarray needs the first dimension name "stack"')
#         #print ('variables', variables)
# 
#         def regression_block(data, variables, weight, valid_pixels_threshold, fit_intercept):
#             data_values  = data.ravel()
#             variables_values = variables.reshape(-1, variables.shape[-1]).T
#             #assert 0, f'TEST: {data_values.shape}, {variables_values.shape}, {weight.shape}'
#             if weight.size > 1:
#                 weight_values = weight.ravel()
#                 nanmask = np.isnan(data_values) | np.isnan(weight_values) | np.any(np.isnan(variables_values), axis=0)
#             else:
#                 weight_values = None
#                 nanmask = np.isnan(data_values) | np.any(np.isnan(variables_values), axis=0)
# 
#             # regression requires enough amount of valid pixels
#             if data_values.size - np.sum(nanmask) < valid_pixels_threshold:
#                 del data_values, variables_values, weight_values, nanmask
#                 return np.nan * np.zeros(data.shape)
# 
#             # build prediction model with or without plane removal (fit_intercept)
#             regr = make_pipeline(StandardScaler(), LinearRegression(fit_intercept=fit_intercept, copy_X=False, n_jobs=1))
#             fit_params = {'linearregression__sample_weight': weight_values[~nanmask]} if weight.size > 1 else {}
#             regr.fit(variables_values[:, ~nanmask].T, data_values[~nanmask], **fit_params)
#             del weight_values, data_values
#             model = np.full_like(data, np.nan).ravel()
#             model[~nanmask] = regr.predict(variables_values[:, ~nanmask].T)
#             del variables_values, regr
#             return model.reshape(data.shape)
# 
#         # xarray wrapper
#         model = xr.apply_ufunc(
#             regression_block,
#             data,
#             variables.chunk(dict(stack=-1, y=chunk2d[0], x=chunk2d[1])),
#             weight.chunk(dict(y=chunk2d[0], x=chunk2d[1])) if weight is not None else weight,
#             dask='parallelized',
#             vectorize=False,
#             output_dtypes=[np.float32],
#             input_core_dims=[[], ['stack'], []],
#             dask_gufunc_kwargs={'valid_pixels_threshold': valid_pixels_threshold, 'fit_intercept': fit_intercept},
#         )
#         return model

    @staticmethod
    def regression(data, variables, weight=None, valid_pixels_threshold=1000, **kwargs):
        """
        topo = sbas.get_topo().coarsen({'x': 4}, boundary='trim').mean()
        yy, xx = xr.broadcast(topo.y, topo.x)
        strat_sbas = sbas.regression(unwrap_sbas.phase,
                [topo,    topo*yy,    topo*xx,    topo*yy*xx,
                 topo**2, topo**2*yy, topo**2*xx, topo**2*yy*xx,
                 topo**3, topo**3*yy, topo**3*xx, topo**3*yy*xx,
                 yy, xx,
                 yy**2, xx**2, yy*xx,
                 yy**3, xx**3, yy**2*xx, xx**2*yy], corr_sbas)


        topo = sbas.interferogram(topophase)
        inc = decimator(sbas.incidence_angle())
        yy, xx = xr.broadcast(topo.y, topo.x)
        variables = [topo,  topo*yy,  topo*xx, topo*yy*xx]
        trend_sbas = sbas.regression(intf_sbas, variables, corr_sbas)
        """
        import numpy as np
        import xarray as xr
        import dask
        # 'linear'
        from sklearn.linear_model import LinearRegression
        # 'sgd'
        from sklearn.linear_model import SGDRegressor
        from sklearn.pipeline import make_pipeline
        from sklearn.preprocessing import StandardScaler

        # find stack dim
        stackvar = data.dims[0] if len(data.dims) >= 3 else 'stack'
        #print ('stackvar', stackvar)
        shape2d = data.shape[1:] if len(data.dims) == 3 else data.shape
        #print ('shape2d', shape2d)
        chunk2d = data.chunks[1:] if len(data.dims) == 3 else data.chunks
        #print ('chunk2d', chunk2d)

        def regression_block(data, weight, *args, **kwargs):
            #assert 0, stack.shape
            data_values  = data.ravel().astype(np.float64)
            # manage variable number of variables
            variables_stack = np.stack([arg[0] if arg.ndim==3 else arg for arg in args])
            #variables_values = variables_stack.reshape(-1, variables_stack.shape[0]).T
            variables_values = variables_stack.reshape(variables_stack.shape[0], -1)
            del variables_stack
            #assert 0, f'TEST: {data_values.shape}, {variables_values.shape}'

            nanmask_data = ~np.isfinite(data_values)
            nanmask_values = np.any(~np.isfinite(variables_values), axis=0)
            if weight.size > 1:
                weight_values = weight.ravel().astype(np.float64)
                nanmask_weight = ~np.isfinite(weight_values)
                nanmask = nanmask_data | nanmask_values | nanmask_weight
                #assert 0, f'TEST weight: {data_values.shape}, {variables_values.shape}, {weight_values.shape}'
            else:
                weight_values = None
                nanmask_weight = None
                nanmask = nanmask_data | nanmask_values

            # regression requires enough amount of valid pixels
            if data_values.size - np.sum(nanmask) < valid_pixels_threshold:
                del data_values, variables_values, weight_values
                del nanmask_data, nanmask_values, nanmask_weight, nanmask
                return np.nan * np.zeros(data.shape)

            # build prediction model with or without plane removal (fit_intercept)
            algorithm = kwargs.pop('algorithm', 'linear')
            if algorithm == 'sgd':
                regr = make_pipeline(StandardScaler(), SGDRegressor(**kwargs))
                fit_params = {'sgdregressor__sample_weight': weight_values[~nanmask]} if weight.size > 1 else {}
            elif algorithm == 'linear':
                regr = make_pipeline(StandardScaler(), LinearRegression(**kwargs, copy_X=False, n_jobs=1))
                fit_params = {'linearregression__sample_weight': weight_values[~nanmask]} if weight.size > 1 else {}
            else:
                raise ValueError(f"Unsupported algorithm {kwargs['algorithm']}. Should be 'linear' or 'sgd'")
            regr.fit(variables_values[:, ~nanmask].T, data_values[~nanmask], **fit_params)
            del weight_values, data_values
            model = np.full_like(data, np.nan).ravel()
            model[~nanmask_values] = regr.predict(variables_values[:, ~nanmask_values].T)
            del variables_values, regr
            del nanmask_data, nanmask_values, nanmask_weight, nanmask
            return model.reshape(data.shape).astype(np.float32)

        dshape = data[0].shape if data.ndim==3 else data.shape
        if isinstance(variables, (list, tuple)):
            vshapes = [v[0].shape if v.ndim==3 else v.shape for v in variables]
            equals = np.all([vshape == dshape for vshape in vshapes])
            assert equals, f'ERROR: shapes of variables slices {vshapes} and data slice {dshape} differ.'
            #assert equals, f'{dshape} {vshapes}, {equals}'
            variables_stack = [v.chunk(dict(y=chunk2d[0], x=chunk2d[1])) for v in variables]
        else:
            vshape = variables[0].shape if variables.ndim==3 else variables.shape
            assert {vshape} == {dshape}, f'ERROR: shapes of variables slice {vshape} and data slice {dshape} differ.'
            variables_stack = [variables.chunk(dict(y=chunk2d[0], x=chunk2d[1]))]

        # xarray wrapper
        model = xr.apply_ufunc(
            regression_block,
            data,
            weight.chunk(dict(y=chunk2d[0], x=chunk2d[1])) if weight is not None else weight,
            *variables_stack,
            dask='parallelized',
            vectorize=False,
            output_dtypes=[np.float32],
            dask_gufunc_kwargs={**kwargs},
        )
        del variables_stack
        return model

    def regression_linear(self, data, variables, weight=None, valid_pixels_threshold=1000, fit_intercept=True):
        """   
        topo = sbas.get_topo().coarsen({'x': 4}, boundary='trim').mean()
        yy, xx = xr.broadcast(topo.y, topo.x)
        strat_sbas = sbas.regression_linear(unwrap_sbas.phase,
                [topo,    topo*yy,    topo*xx,    topo*yy*xx,
                 topo**2, topo**2*yy, topo**2*xx, topo**2*yy*xx,
                 topo**3, topo**3*yy, topo**3*xx, topo**3*yy*xx,
                 yy, xx,
                 yy**2, xx**2, yy*xx,
                 yy**3, xx**3, yy**2*xx, xx**2*yy], corr_sbas)
        """
        return self.regression(data, variables, weight, valid_pixels_threshold, algorithm='linear',
                                fit_intercept=fit_intercept)

    def regression_sgd(self, data, variables, weight=None, valid_pixels_threshold=1000,
                      max_iter=1000, tol=1e-3, alpha=0.0001, l1_ratio=0.15):
        """
        Perform regression on a dataset using the SGDRegressor model from scikit-learn.

        This function applies Stochastic Gradient Descent (SGD) regression to fit the given 'data' against a set of 'variables'. It's suitable for large datasets and handles high-dimensional features efficiently.

        Parameters:
        data (xarray.DataArray): The target data array to fit.
        variables (xarray.DataArray or list of xarray.DataArray): Predictor variables. It can be a single 3D DataArray or a list of 2D DataArrays.
        weight (xarray.DataArray, optional): Weights for each data point. Useful if certain data points are more important. Defaults to None.
        valid_pixels_threshold (int, optional): Minimum number of valid pixels required for the regression to be performed. Defaults to 10000.
        max_iter (int, optional): Maximum number of passes over the training data (epochs). Defaults to 1000.
        tol (float, optional): Stopping criterion. If not None, iterations will stop when (loss > best_loss - tol) for n_iter_no_change consecutive epochs. Defaults to 1e-3.
        alpha (float, optional): Constant that multiplies the regularization term. Higher values mean stronger regularization. Defaults to 0.0001.
        l1_ratio (float, optional): The Elastic Net mixing parameter, with 0 <= l1_ratio <= 1. l1_ratio=0 corresponds to L2 penalty, l1_ratio=1 to L1. Defaults to 0.15.

        Returns:
        xarray.DataArray: The predicted values as an xarray DataArray, fitted by the SGDRegressor model.

        Notes:
        - SGDRegressor is well-suited for large datasets due to its efficiency in handling large-scale and high-dimensional data.
        - Proper tuning of parameters (max_iter, tol, alpha, l1_ratio) is crucial for optimal performance and prevention of overfitting.

        Example:  
        decimator = sbas.decimator(resolution=15, grid=(1,1))
        topo = decimator(sbas.get_topo())
        inc = decimator(sbas.incidence_angle())
        yy, xx = xr.broadcast(topo.y, topo.x)
        trend_sbas = sbas.regression(unwrap_sbas.phase,
                [topo,    topo*yy,    topo*xx,    topo*yy*xx,
                 topo**2, topo**2*yy, topo**2*xx, topo**2*yy*xx,
                 topo**3, topo**3*yy, topo**3*xx, topo**3*yy*xx,
                 inc,     inc**yy,    inc*xx,     inc*yy*xx,
                 yy, xx,
                 yy**2, xx**2, yy*xx,
                 yy**3, xx**3, yy**2*xx, xx**2*yy], corr_sbas)
        """
        return self.regression(data, variables, weight, valid_pixels_threshold, algorithm='sgd',
                               max_iter=max_iter, tol=tol, alpha=alpha, l1_ratio=l1_ratio)

    def polyfit(self, data, weight=None, degree=0, days=None, count=None):
        import xarray as xr
        import pandas as pd
        import numpy as np

        pairs, dates = self.get_pairs(data, dates=True)

        models = []
        for date in dates:
            #print ('date', date)
            data_pairs = pairs[(pairs.ref==date)|(pairs.rep==date)].pair.values
            #print ('data_pairs', data_pairs)
            if weight is None:
                stack = data.sel(pair=data_pairs)
            else:
                stack = data.sel(pair=data_pairs) * np.sqrt(weight.sel(pair=data_pairs))
            stack_days = xr.where(stack.ref < pd.Timestamp(date),
                           (stack.ref - stack.rep).dt.days,
                           (stack.rep - stack.ref).dt.days)
            #print ('stack_days', stack_days)
            # select smallest intervals
            stack_days_selected = stack_days[np.argsort(np.abs(stack_days.values))][:count]
            if days is not None:
                stack_days_selected = stack_days_selected[np.abs(stack_days_selected)<=days]
            #print ('stack_days_selected', stack_days_selected)
            linear_fit = (np.sign(stack_days)*stack).assign_coords(time=stack_days)\
                [stack.pair.isin(stack_days_selected.pair)]\
                .swap_dims({'pair': 'time'})\
                .sortby(['ref', 'rep'])\
                .polyfit(dim='time', deg=degree)
            model = linear_fit.polyfit_coefficients.sel(degree=degree)
            models.append(model.assign_coords(date=pd.to_datetime(date)))
            del data_pairs, stack, stack_days, stack_days_selected, linear_fit, model
        model = xr.concat(models, dim='date')
        del models

        out = xr.concat([(model.sel(date=ref).drop('date') - model.sel(date=rep).drop('date'))\
                                 .assign_coords(pair=str(ref.date()) + ' ' + str(rep.date()), ref=ref, rep=rep) \
                          for ref, rep in zip(pairs['ref'], pairs['rep'])], dim='pair')
        return out.rename(data.name)

#     def polyfit(self, data, weight=None, degree=0, variable=None, count=None):
#         import xarray as xr
#         import pandas as pd
#         import numpy as np
# 
#         pairs, dates = self.get_pairs(data, dates=True)
#         if variable is not None:
#             pairs['variable'] = variable.values if isinstance(variable, pd.Series) else variable
# 
#         models = []
#         for date in dates:
#             data_pairs = pairs[(pairs.ref==date)|(pairs.rep==date)].pair.values
#             #print (data_pairs)
#             stack = data.sel(pair=data_pairs)
#             if weight is None:
#                 stack = data.sel(pair=data_pairs)
#             else:
#                 stack = data.sel(pair=data_pairs) * np.sqrt(weight.sel(pair=data_pairs))
#             # days, positive and negative
#             days = xr.where(stack.ref < pd.Timestamp(date),
#                            (stack.ref - stack.rep).dt.days,
#                            (stack.rep - stack.ref).dt.days)
#             # select smallest intervals
#             days_selected = days[np.argsort(np.abs(days.values))][:count]
#             if variable is not None:
#                 data_variables = xr.DataArray(pairs[pairs.pair.isin(data_pairs)]['variable'].values, coords=days.coords)
#             else:
#                 data_variables = days
#             #print (days, days_selected, data_variables)
#             #.sortby(['ref', 'rep'])
#             linear_fit = (np.sign(days)*stack)\
#                 .assign_coords(time=data_variables)[stack.pair.isin(days_selected.pair)]\
#                 .swap_dims({'pair': 'variable'})\
#                 .polyfit(dim='variable', deg=degree)
#             model = linear_fit.polyfit_coefficients.sel(degree=degree)
#             models.append(model.assign_coords(date=pd.to_datetime(date)))
#             del data_pairs, stack, days, days_selected, data_variables, linear_fit, model
#         model = xr.concat(models, dim='date')
#         del models
# 
#         out = xr.concat([(model.sel(date=ref).drop('date') - model.sel(date=rep).drop('date'))\
#                                  .assign_coords(pair=str(ref.date()) + ' ' + str(rep.date()), ref=ref, rep=rep) \
#                           for ref, rep in zip(pairs['ref'], pairs['rep'])], dim='pair')
# 
#         return out.rename(data.name)

    def gaussian(self, grid, wavelength, truncate=3.0, resolution=60, debug=False):
        """
        Apply a lazy Gaussian filter to an input 2D or 3D data array.

        Parameters
        ----------
        dataarray : xarray.DataArray
            The input data array with NaN values allowed.
        wavelength : float
            The cut-off wavelength for the Gaussian filter in meters.
        truncate : float, optional
            Size of the Gaussian kernel, defined in terms of standard deviation, or 'sigma'. 
            It is the number of sigmas at which the window (filter) is truncated. 
            For example, if truncate = 3.0, the window will cut off at 3 sigma. Default is 3.0.
        resolution : float, optional
            The processing resolution for the Gaussian filter in meters.
        debug : bool, optional
            Whether to print debug information.

        Returns
        -------
        xarray.DataArray
            The filtered data array with the same coordinates as the input.

        Examples
        --------
        Detrend ionospheric effects and solid Earth's tides on a large area and save to disk:
        stack.stack_gaussian2d(slcs, wavelength=400)
        For band-pass filtering apply the function twice and save to disk:
        model = stack.stack_gaussian2d(slcs, wavelength=400, interactive=True) \
            - stack.stack_gaussian2d(slcs, wavelength=2000, interactive=True)
        stack.save_cube(model, caption='Gaussian Band-Pass filtering')

        Detrend and return lazy xarray dataarray:
        stack.stack_gaussian2d(slcs, wavelength=400, interactive=True)
        For band-pass filtering apply the function twice:
        stack.stack_gaussian2d(slcs, wavelength=400, interactive=True) \
            - stack.stack_gaussian2d(slcs, wavelength=2000, interactive=True) 

        """
        import xarray as xr
        import numpy as np
#         import warnings
#         # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
#         warnings.filterwarnings('ignore')
#         warnings.filterwarnings('ignore', module='dask')
#         warnings.filterwarnings('ignore', module='dask.core')

        assert self.is_ra(grid), 'ERROR: the processing requires grid in radar coordinates'
        assert np.issubdtype(grid.dtype, np.floating), 'ERROR: expected float datatype input grid'
        assert wavelength is not None, 'ERROR: Gaussian filter cut-off wavelength is not defined'

        # ground pixel size
        dy, dx = self.get_spacing(grid)
        # downscaling
        yscale, xscale = int(np.round(resolution/dy)), int(np.round(resolution/dx))
        # gaussian kernel
        #sigma_y = np.round(wavelength / dy / yscale, 1)
        #sigma_x = np.round(wavelength / dx / xscale, 1)
        if debug:
            print (f'DEBUG: gaussian: ground pixel size in meters: y={dy:.1f}, x={dx:.1f}')
        if (xscale <=1 and yscale <=1) or (wavelength/resolution <= 3):
            # decimation is useless
            return self.multilooking(grid, wavelength=wavelength, coarsen=None, debug=debug)

        # define filter on decimated grid, the correction value is typically small
        wavelength_dec = np.sqrt(wavelength**2 - resolution**2)
        if debug:
            print (f'DEBUG: gaussian: downscaling to resolution {resolution}m using yscale {yscale}, xscale {xscale}')
            #print (f'DEBUG: gaussian: filtering on {resolution}m grid using sigma_y0 {sigma_y}, sigma_x0 {sigma_x}')
            print (f'DEBUG: gaussian: filtering on {resolution}m grid using wavelength {wavelength_dec:.1f}')

        # find stack dim
        stackvar = grid.dims[0] if len(grid.dims) == 3 else 'stack'
        #print ('stackvar', stackvar)

        # split coordinate grid to equal chunks and rest
        ys_blocks = np.array_split(grid.y, np.arange(0, grid.y.size, self.chunksize)[1:])
        xs_blocks = np.array_split(grid.x, np.arange(0, grid.x.size, self.chunksize)[1:])

        grid_dec = self.multilooking(grid, wavelength=resolution, coarsen=(yscale,xscale), debug=debug)
        grid_dec_gauss = self.multilooking(grid_dec, wavelength=wavelength_dec, debug=debug)
        del grid_dec
    
        stack = []
        for stackval in grid[stackvar].values if len(grid.dims) == 3 else [None]:
            grid_in = grid_dec_gauss.sel({stackvar: stackval}) if stackval is not None else grid_dec_gauss
            grid_out = grid_in.reindex({'y': grid.y, 'x': grid.x}, method='nearest').chunk(self.chunksize)
            del grid_in
            stack.append(grid_out)
            del grid_out

        # wrap lazy Dask array to Xarray dataarray
        if len(grid.dims) == 2:
            out = stack[0]
        else:
            out = xr.concat(stack, dim=stackvar)
        del stack

        # append source grid coordinates excluding removed y, x ones
        for (k,v) in grid.coords.items():
            if k not in ['y','x']:
                out[k] = v

        return out

#     def detrend(self, dataarray, fit_intercept=True, fit_dem=True, fit_coords=True,
#                 resolution=90, debug=False):
#         """
#         Detrend and return output for a single unwrapped interferogram combining optional topography and linear components removal.
# 
#         Parameters
#         ----------
#         dataarray : xarray.DataArray
#             The input data array to detrend.
#         fit_intercept : bool, optional
#             Whether to remove the mean value (plane) from the data. Default is True.
#         fit_dem : bool, optional
#             Whether to detrend the topography. Default is True.
#         fit_coords : bool, optional
#             Whether to detrend the linear coordinate components. Default is True.
#         resolution : int, optional
#             The processing resolution to prevent overfitting and reduce grid size. Default is 90.
#         debug : bool, optional
#             Whether to print debug information. Default is False.
# 
#         Returns
#         -------
#         xarray.DataArray
#             The detrended 2D data array.
# 
#         Examples
#         --------
#         Simplest detrending:
#         unwrap_detrended = stack.detrend(pair.values[0] if isinstance(pairs, pd.DataFrame) else pair[0])
# 
#         Detrend unwrapped interferogram in radar coordinates, see for details:
#         - [GitHub Issue 98](https://github.com/gmtsar/gmtsar/issues/98)
#         - [GitHub Issue 411](https://github.com/gmtsar/gmtsar/issues/411)
#         """
#         import xarray as xr
#         import numpy as np
#         import dask
#         from sklearn.linear_model import LinearRegression
#         from sklearn.pipeline import make_pipeline
#         from sklearn.preprocessing import StandardScaler
# 
#         def postprocessing(out):
#             return out.astype(np.float32).rename('detrend')
# 
#         # check the simplest case
#         if not fit_intercept and not fit_dem and not fit_coords:
#             print ('NOTE: All the detrending options disable, function does nothing')
#             return dataarray
# 
#         # check simple case
#         if fit_intercept and not fit_dem and not fit_coords:
#             if debug:
#                 print ('DEBUG: Remove mean value only')
#             return postprocessing(dataarray - dataarray.mean())
# 
#         # input grid can be too large
#         decimator = self.pixel_decimator(resolution=resolution, grid=dataarray, debug=debug)
#         # decimate
#         dataarray_dec = decimator(dataarray)
#         if debug:
#             print ('DEBUG: Decimated data array', dataarray_dec.shape)
# 
#         # topography grid required to fit_dem option only
#         if fit_dem:
#             if debug:
#                 print ('DEBUG: Interpolate topography on the data grid')
#             topo = self.get_topo()
#             #topo = topo.reindex_like(unwraps[0], method='nearest')
#             # use xr.zeros_like to prevent the target grid coordinates modifying
#             topo = topo.reindex_like(dataarray, method='nearest')
#             # check chunks
#             if debug:
#                 print ('DEBUG: regrid to resolution in meters', resolution)
#             # decimate
#             topo_dec  = decimator(topo)
#             if debug:
#                 print ('DEBUG: Decimated topography array', topo_dec.shape)
#         else:
#             topo = topo_dec = None
# 
#         # lazy calculations are useless below
#         def data2fit(data, elev, yy, xx):
#             y = data.values.reshape(-1) if isinstance(data, xr.DataArray) else data.reshape(-1)
#             nanmask = np.isnan(y)
#             # prepare regression variable
#             Y = y[~nanmask]
# 
#             if fit_coords or fit_dem:
#                 # prepare coordinates for X regression variable
#                 ys = (yy.values.reshape(-1) if isinstance(yy, xr.DataArray) else yy.reshape(-1))[~nanmask]
#                 xs = (xx.values.reshape(-1) if isinstance(xx, xr.DataArray) else xx.reshape(-1))[~nanmask]
# 
#             if fit_dem:
#                 # prepare topography for X regression variable
#                 zs = (elev.values.reshape(-1) if isinstance(elev, xr.DataArray) else elev.reshape(-1))[~nanmask]
#                 zys = zs*ys
#                 zxs = zs*xs
# 
#             if fit_dem and fit_coords:
#                 X = np.column_stack([zys, zxs, ys, xs, zs])
#             elif fit_dem:
#                 X = np.column_stack([zys, zxs, zs])
#             elif fit_coords:
#                 X = np.column_stack([ys, xs])
#             return Y, X, nanmask
# 
#         if debug:
#             print ('DEBUG: linear regression calculation')
# 
#         def regr_fit():
#             # build prediction model with or without plane removal (fit_intercept)
#             regr = make_pipeline(StandardScaler(), LinearRegression(fit_intercept=fit_intercept))
#             yy, xx = xr.broadcast(dataarray_dec.y, dataarray_dec.x)
#             Y, X, _ = data2fit(dataarray_dec, topo_dec, yy, xx)
# 
#             return regr.fit(X, Y)
# 
#         # calculate for chunks
#         def predict(data, elev, yy, xx, regr):
#             Y, X, nanmask = data2fit(data, elev, yy, xx)
#             # the chunk is NaN-filled, prediction impossible
#             if nanmask.all():
#                 return data
#             # predict when some values are not NaNs
#             model = np.nan * np.zeros(data.shape)
#             model.reshape(-1)[~nanmask] = regr.predict(X)
#             # return data without the trend
#             return data - model
# 
#         def regr_predict(regr):
#             yy = xr.DataArray(dataarray.y).chunk(-1)
#             xx = xr.DataArray(dataarray.x).chunk(-1)
#             yy, xx = xr.broadcast(yy, xx)
# 
#             # xarray wrapper
#             return xr.apply_ufunc(
#                 predict,
#                 dataarray,
#                 topo.chunk(dataarray.chunks) if topo is not None else None,
#                 yy.chunk(dataarray.chunks),
#                 xx.chunk(dataarray.chunks),
#                 dask='parallelized',
#                 vectorize=False,
#                 output_dtypes=[np.float32],
#                 dask_gufunc_kwargs={'regr': regr},
#             )
# 
#         # build the model and return the input data without the detected trend
#         return postprocessing(regr_predict(regr_fit()))

#     def gaussian(self, grid, wavelength, truncate=3.0, resolution=90, debug=False):
#         """
#         Apply a lazy Gaussian filter to an input 2D or 3D data array.
# 
#         Parameters
#         ----------
#         dataarray : xarray.DataArray
#             The input data array with NaN values allowed.
#         wavelength : float
#             The cut-off wavelength for the Gaussian filter in meters.
#         truncate : float, optional
#             Size of the Gaussian kernel, defined in terms of standard deviation, or 'sigma'. 
#             It is the number of sigmas at which the window (filter) is truncated. 
#             For example, if truncate = 3.0, the window will cut off at 3 sigma. Default is 3.0.
#         resolution : float, optional
#             The processing resolution for the Gaussian filter in meters.
#         debug : bool, optional
#             Whether to print debug information.
# 
#         Returns
#         -------
#         xarray.DataArray
#             The filtered data array with the same coordinates as the input.
# 
#         Examples
#         --------
#         Detrend ionospheric effects and solid Earth's tides on a large area and save to disk:
#         stack.stack_gaussian2d(slcs, wavelength=400)
#         For band-pass filtering apply the function twice and save to disk:
#         model = stack.stack_gaussian2d(slcs, wavelength=400, interactive=True) \
#             - stack.stack_gaussian2d(slcs, wavelength=2000, interactive=True)
#         stack.save_cube(model, caption='Gaussian Band-Pass filtering')
# 
#         Detrend and return lazy xarray dataarray:
#         stack.stack_gaussian2d(slcs, wavelength=400, interactive=True)
#         For band-pass filtering apply the function twice:
#         stack.stack_gaussian2d(slcs, wavelength=400, interactive=True) \
#             - stack.stack_gaussian2d(slcs, wavelength=2000, interactive=True) 
# 
#         """
#         import xarray as xr
#         import numpy as np
# 
#         assert self.is_ra(grid), 'ERROR: the processing requires grid in radar coordinates'
#         assert wavelength is not None, 'ERROR: Gaussian filter cut-off wavelength is not defined'
# 
#         # ground pixel size
#         dy, dx = self.get_spacing(grid)
#         # reduction
#         yscale, xscale = int(np.round(resolution/dy)), int(np.round(resolution/dx))
#         # gaussian kernel
#         sigma_y = np.round(wavelength / dy / yscale)
#         sigma_x = np.round(wavelength / dx / xscale)
#         if debug:
#             print (f'DEBUG: average ground pixel size in meters: y={dy}, x={dx}')
#             print (f'DEBUG: yscale {yscale}, xscale {xscale} to resolution {resolution} m')
#             print ('DEBUG: Gaussian filtering using resolution, sigma_y, sigma_x', resolution, sigma_y, sigma_x)
# 
#         # find stack dim
#         stackvar = grid.dims[0] if len(grid.dims) == 3 else 'stack'
#         #print ('stackvar', stackvar)
#     
#         stack = []
#         for stackval in grid[stackvar].values if len(grid.dims) == 3 else [None]:
#             block = grid.sel({stackvar: stackval}) if stackval is not None else grid
#             block_dec = self.antialiasing_downscale(block, wavelength=resolution, coarsen=(yscale,xscale), debug=debug)
#             gaussian_dec = self.nanconvolve2d_gaussian(block_dec, (sigma_y,sigma_x), truncate=truncate)
#             # interpolate decimated filtered grid to original resolution
#             gaussian = gaussian_dec.interp_like(block, method='nearest')
#             # revert the original chunks
#             gaussian = xr.unify_chunks(block, gaussian)[1]
#             stack.append(gaussian.astype(np.float32).rename('gaussian'))
# 
#         # wrap lazy Dask array to Xarray dataarray
#         if len(grid.dims) == 2:
#             out = stack[0]
#         else:
#             out = xr.concat(stack, dim=stackvar)
#         del stack
# 
#         # append source grid coordinates excluding removed y, x ones
#         for (k,v) in grid.coords.items():
#             if k not in ['y','x']:
#                 out[k] = v
#     
#         return out

#     def turbulence(self, phase, date_crop=0, symmetrical=False):
#         pairs, dates = self.get_pairs(phase, dates=True)
#     
#         turbos = []
#         for date in dates:
#             ref = pairs[pairs.ref==date]
#             rep = pairs[pairs.rep==date]
#             # calculate left and right pairs
#             if symmetrical:
#                 count = min(len(ref), len(rep))
#             else:
#                 count = None
#             #print (date, len(ref), len(rep), '=>', count)
#             if len(ref) < 1 or len(rep) < 1:
#                 # correction calculation is not possible
#                 #turbo = xr.zeros_like((detrend - phase_topo).isel(pair=0)).drop(['pair','ref','rep'])
#                 continue
#             else:
#                 ref_data = phase.sel(pair=ref.pair.values).isel(pair=slice(None,count))
#                 #ref_weight = 1/corr60m.sel(pair=ref_data.pair)
#                 #print (ref_data)
#                 rep_data = phase.sel(pair=rep.pair.values ).isel(pair=slice(None,count))
#                 #rep_weight = 1/corr60m.sel(pair=rep_data.pair)
#                 #print (rep_data)
#                 #mask = (ref_weight.sum('pair') + rep_weight.sum('pair'))/2/count
#                 turbo = (ref_data.mean('pair') - rep_data.mean('pair')) / 2
#                 #turbo = ((ref_data*ref_weight).mean('pair')/ref_weight.sum('pair') -\
#                 #         (rep_data*rep_weight).mean('pair')/rep_weight.sum('pair')) / 2
#                 #.where(mask>MASK)
#             turbos.append(turbo.assign_coords(date=pd.to_datetime(date)))
#             del turbo
#         turbo = xr.concat(turbos, dim='date')
#         del turbos
#     
#         # convert dates to pairs
#         #pairs = pairs[pairs.ref.isin(turbo.date.values)&pairs.rep.isin(turbo.date.values)]
#         fake = xr.zeros_like(phase.isel(pair=0))
# 
#     #     return [(
#     #         (turbo.sel(date=ref).drop('date') if ref in turbo.date else fake) - \
#     #         (turbo.sel(date=rep).drop('date') if rep in turbo.date else fake)
#     #     ).assign_coords(pair=str(ref.date()) + ' ' + str(rep.date()), ref=ref, rep=rep) \
#     #     for ref, rep in zip(pairs['ref'], pairs['rep'])]
#     
#     
#         dates_crop = dates[date_crop:None if date_crop is None or date_crop==0 else -date_crop]
#         pairs_crop = pairs[pairs.ref.isin(dates_crop)&pairs.rep.isin(dates_crop)]
#     
#     #     refs = [ref for ref in pairs['ref'] if str(ref.date()) in dates]
#     #     reps = [rep for rep in pairs['rep'] if str(rep.date()) in dates]
#     
#         phase_turbo = xr.concat([(
#             (turbo.sel(date=ref).drop('date') if ref in turbo.date else fake) - \
#             (turbo.sel(date=rep).drop('date') if rep in turbo.date else fake)
#         ).assign_coords(pair=str(ref.date()) + ' ' + str(rep.date()), ref=ref, rep=rep) \
#         for ref, rep in zip(pairs_crop['ref'], pairs_crop['rep'])], dim='pair')
#     
#     #     phase_turbo = xr.concat([(turbo.sel(date=ref) - turbo.sel(date=rep))\
#     #                              .assign_coords(pair=str(ref.date()) + ' ' + str(rep.date()), ref=ref, rep=rep) \
#     #                              for ref, rep in \
#     #                              zip(pairs['ref'], pairs['rep'])],
#     #                        dim='pair')
#         #phase_turbo['ref'].values = pd.to_datetime(phase_turbo['ref'])
#         #phase_turbo['rep'].values = pd.to_datetime(phase_turbo['rep'])
#         return phase_turbo

#     def turbulence(self, phase, weight=None, date_crop=1, symmetrical=False):
#         import xarray as xr
#         import pandas as pd
# 
#         pairs, dates = self.get_pairs(phase, dates=True)
# 
#         turbos = []
#         for date in dates:
#             ref = pairs[pairs.ref==date]
#             rep = pairs[pairs.rep==date]
#             # calculate left and right pairs
#             count = min(len(ref), len(rep)) if symmetrical else None
#             #print (date, len(ref), len(rep), '=>', count)
#             if len(ref) < 1 or len(rep) < 1:
#                 # correction calculation is not possible
#                 continue
#             ref_data = phase.sel(pair=ref.pair.values).isel(pair=slice(None,count))
#             #print (ref_data)
#             rep_data = phase.sel(pair=rep.pair.values ).isel(pair=slice(None,count))
#             #print (rep_data)
#             if weight is not None:
#                 ref_weight = weight.sel(pair=ref_data.pair)
#                 rep_weight = weight.sel(pair=rep_data.pair)
#                 turbo = ((ref_data*ref_weight).mean('pair')/ref_weight.sum('pair') -\
#                          (rep_data*rep_weight).mean('pair')/rep_weight.sum('pair')) / 2
#                 del ref_weight, rep_weight
#             else:
#                 turbo = (ref_data.mean('pair') - rep_data.mean('pair')) / 2
#             del ref_data, rep_data
#             turbos.append(turbo.assign_coords(date=pd.to_datetime(date)))
#             del turbo
#         turbo = xr.concat(turbos, dim='date')
#         del turbos
# 
#         # empty grid
#         empty = xr.zeros_like(phase.isel(pair=0))
#         # convert dates to pairs
#         dates_crop = dates[date_crop:None if date_crop is None or date_crop==0 else -date_crop]
#         pairs_crop = pairs[pairs.ref.isin(dates_crop) & pairs.rep.isin(dates_crop)]  
#         phase_turbo = xr.concat([(
#             (turbo.sel(date=ref).drop('date') if ref in turbo.date else empty) - \
#             (turbo.sel(date=rep).drop('date') if rep in turbo.date else empty)
#         ).assign_coords(pair=str(ref.date()) + ' ' + str(rep.date()), ref=ref, rep=rep) \
#         for ref, rep in zip(pairs_crop['ref'], pairs_crop['rep'])], dim='pair')
#         del empty, dates_crop, pairs_crop
# 
#         return phase_turbo.rename('turbulence')

    def turbulence(self, phase, weight=None):
        import xarray as xr
        import pandas as pd

        pairs, dates = self.get_pairs(phase, dates=True)

        turbos = []
        for date in dates:
            ref = pairs[pairs.ref==date]
            rep = pairs[pairs.rep==date]
            #print (date, len(ref), len(rep))
            ref_data = phase.sel(pair=ref.pair.values)
            #print (ref_data)
            rep_data = phase.sel(pair=rep.pair.values)
            #print (rep_data)
            if weight is not None:
                ref_weight = weight.sel(pair=ref.pair.values)
                rep_weight = weight.sel(pair=rep.pair.values)
                turbo = xr.concat([ref_data*ref_weight, -rep_data*rep_weight], dim='pair').sum('pair')/\
                    xr.concat([ref_weight, rep_weight], dim='pair').sum('pair')
                del ref_weight, rep_weight
            else:
                turbo = xr.concat([ref_data, -rep_data], dim='pair').mean('pair')
            del ref_data, rep_data
            turbos.append(turbo.assign_coords(date=pd.to_datetime(date)))
            del turbo
        turbo = xr.concat(turbos, dim='date')
        del turbos

        phase_turbo = xr.concat([(turbo.sel(date=ref).drop('date') - turbo.sel(date=rep).drop('date'))\
                                 .assign_coords(pair=str(ref.date()) + ' ' + str(rep.date()), ref=ref, rep=rep) \
                          for ref, rep in zip(pairs['ref'], pairs['rep'])], dim='pair')

        return phase_turbo.rename('turbulence')



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_phasediff import Stack_phasediff

class Stack_multilooking(Stack_phasediff):

    #decimator = lambda da: da.coarsen({'y': 2, 'x': 2}, boundary='trim').mean()
    def decimator(self, resolution=60, grid=(1, 4), func='mean', debug=False):
        """
        Return function for pixel decimation to the specified output resolution.

        Parameters
        ----------
        resolution : int, optional
            DEM grid resolution in meters. The same grid is used for geocoded results output.
        grid : tuple, optional
            Grid size for pixel decimation in the format (vertical, horizontal).
        debug : bool, optional
            Boolean flag to print debug information.

        Returns
        -------
        callable
            Post-processing lambda function.

        Examples
        --------
        Decimate computed interferograms to default DEM resolution 60 meters:
        decimator = stack.decimator()
        stack.intf(pairs, func=decimator)
        """
        import numpy as np
        import dask
        import warnings
        # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
        warnings.filterwarnings('ignore')
        warnings.filterwarnings('ignore', module='dask')
        warnings.filterwarnings('ignore', module='dask.core')

        # special cases: scale factor should be 4*N or 2*N to prevent rounding issues
        # grid can be defined as [] or () or xarray dataarray
        grid_as_coeffs = isinstance(grid,(list, tuple))
        if   (grid_as_coeffs and grid==(1, 1)) or (not grid_as_coeffs and grid.x.diff('x')[0].item()) == 1:
            xscale0 = 4
        elif (grid_as_coeffs and grid==(1, 2)) or (not grid_as_coeffs and grid.x.diff('x')[0].item()) == 2:
            xscale0 = 2
        else:
            xscale0 = 1
        if debug:
            print (f'DEBUG: scale to square grid: xscale0={xscale0}')

        dy, dx = self.get_spacing(grid)
        yscale, xscale = int(np.round(resolution/dy)), int(np.round(resolution/dx/xscale0))
        if debug:
            print (f'DEBUG: ground pixel size in meters: y={dy:.1f}, x={dx:.1f}')
        if yscale <= 1 and xscale <= 1 and xscale0==1:
            # decimation impossible
            if debug:
                print (f'DEBUG: decimator = lambda da: da')
            return lambda da: da
        if debug:
            print (f"DEBUG: decimator = lambda da: da.coarsen({{'y': {yscale}, 'x': {xscale0*xscale}}}, boundary='trim').{func}()")

        # decimate function
        def decimator(da):
            import warnings
            # suppress Dask warning "RuntimeWarning: invalid value encountered in divide"
            warnings.filterwarnings('ignore')
            warnings.filterwarnings('ignore', module='dask')
            warnings.filterwarnings('ignore', module='dask.core')
            # workaround for Google Colab when we cannot save grids with x,y coordinate names
            # also supports geographic coordinates
            yname = [varname for varname in ['y', 'lat', 'a'] if varname in da.dims][0]
            xname = [varname for varname in ['x', 'lon', 'r'] if varname in da.dims][0]
            coarsen_args = {yname: yscale, xname: xscale0*xscale}
            # avoid creating the large chunks
            with dask.config.set(**{'array.slicing.split_large_chunks': True}):
                if func not in ['mean', 'min', 'max', 'count', 'sum']:
                    raise ValueError(f"Unsupported function {func}. Should be 'mean','min','max','count', or 'sum'")
                return getattr(da.coarsen(coarsen_args, boundary='trim'), func)()\
                       .chunk({yname: self.chunksize, xname: self.chunksize})

        # return callback function and set common chunk size
        return lambda da: decimator(da)

    # coarsen = None disables downscaling and uses wavelength to filter
    # coarsen=1 disables downscaling and use coarsen/cutoff filter
    def multilooking(self, data, weight=None, wavelength=None, coarsen=None, debug=False):
        import xarray as xr
        import numpy as np
        import dask
        from dask_image.ndfilters import gaussian_filter as dask_gaussian_filter
        # GMTSAR constant 5.3 defines half-gain at filter_wavelength
        # https://github.com/gmtsar/gmtsar/issues/411
        cutoff = 5.3

        # expand simplified definition
        if coarsen is not None and not isinstance(coarsen, (list, tuple, np.ndarray)):
            coarsen = (coarsen, coarsen)

        # allow this case to save the original grid resolution
        if wavelength is None and coarsen is None:
            return data

        # antialiasing (multi-looking) filter
        if wavelength is None:
            sigmas = [coarsen[0]/cutoff, coarsen[1]/cutoff]
            if debug:
                print (f'DEBUG: multilooking sigmas ({sigmas[0]:.2f}, {sigmas[1]:.2f}), for specified coarsen {coarsen}')
        else:
            dy, dx = self.get_spacing(data)
            #print ('DEBUG dy, dx', dy, dx)
            #sigmas = int(np.round(wavelength/dy/coarsen[0])), int(np.round(wavelength/dx))
            sigmas = [wavelength/cutoff/dy, wavelength/cutoff/dx]
            if debug:
                print (f'DEBUG: multilooking sigmas ({sigmas[0]:.2f}, {sigmas[1]:.2f}), for specified wavelength {wavelength:.1f}')

        # weighted and not weighted convolution on float and complex float data
        def apply_filter(data, weight, sigmas, truncate=2):
            if np.issubdtype(data.dtype, np.complexfloating):
                #print ('complexfloating')
                parts = []
                for part in [data.real, data.imag]:
                    data_complex  = ((1j + part) * (weight if weight is not None else 1)).fillna(0)
                    conv_complex = dask_gaussian_filter(data_complex.data, sigmas, mode='reflect', truncate=truncate)
                    #conv = conv_complex.real/conv_complex.imag
                    # to prevent "RuntimeWarning: invalid value encountered in divide" even when warning filter is defined
                    conv = dask.array.where(conv_complex.imag == 0, np.nan, conv_complex.real/(conv_complex.imag + 1e-17))
                    del data_complex, conv_complex
                    parts.append(conv)
                    del conv
                conv = parts[0] + 1j*parts[1]
                del parts
            else:
                #print ('floating')
                # replace nan + 1j to to 0.+0.j
                data_complex  = ((1j + data) * (weight if weight is not None else 1)).fillna(0)
                conv_complex = dask_gaussian_filter(data_complex.data, sigmas, mode='reflect', truncate=truncate)
                #conv = conv_complex.real/conv_complex.imag
                # to prevent "RuntimeWarning: invalid value encountered in divide" even when warning filter is defined
                conv = dask.array.where(conv_complex.imag == 0, np.nan, conv_complex.real/(conv_complex.imag + 1e-17))
                del data_complex, conv_complex
            return conv

        if isinstance(data, xr.Dataset):
            dims = data[list(data.data_vars)[0]].dims
        else:
            dims = data.dims

        if len(dims) == 2:
            stackvar = None
        else:
            stackvar = dims[0]
        #print ('stackvar', stackvar)

        if weight is not None:
            # for InSAR processing expect 2D weights
            assert isinstance(weight, xr.DataArray) and len(weight.dims)==2, \
                'ERROR: multilooking weight should be 2D DataArray'
        
        if weight is not None and len(data.dims) == len(weight.dims):
            #print ('2D check shape weighted')
            # single 2D grid processing
            if isinstance(data, xr.Dataset):
                for varname in data.data_vars:
                    assert data[varname].shape == weight.shape, \
                        f'ERROR: multilooking data[{varname}] and weight variables have different shape'
            else:
                assert data.shape == weight.shape, 'ERROR: multilooking data and weight variables have different shape'
        elif weight is not None and len(data.dims) == len(weight.dims) + 1:
            #print ('3D check shape weighted')
            # stack of 2D grids processing
            if isinstance(data, xr.Dataset):
                for varname in data.data_vars:
                    assert data[varname].shape[1:] == weight.shape, \
                        f'ERROR: multilooking data[{varname}] slice and weight variables have different shape'
            else:
                assert data.shape[1:] == weight.shape, 'ERROR: multilooking data slice and weight variables have different shape'

        stack =[]
        for ind in range(len(data[stackvar]) if stackvar is not None else 1):
            if isinstance(data, xr.Dataset):
                #print (f'Dataset ind:{ind}')
                data_convs = []
                for key in data.data_vars:
                    conv = apply_filter(data[key][ind] if stackvar is not None else data[key],
                                        weight,
                                        sigmas)
                    data_conv = xr.DataArray(conv, dims=data[key].dims[1:] if stackvar is not None else data[key].dims, name=data[key].name)
                    del conv
                    data_convs.append(data_conv)
                    del data_conv
                stack.append(xr.merge(data_convs))
                del data_convs
            else:
                #print (f'DataArray ind:{ind}')
                conv = apply_filter(data[ind]   if stackvar is not None else data,
                                    weight,
                                    sigmas)
                data_conv = xr.DataArray(conv, dims=data.dims[1:] if stackvar is not None else data.dims, name=data.name)
                del conv
                stack.append(data_conv)
                del data_conv

        if stackvar is not None:
            #print ('3D')
            ds = xr.concat(stack, dim=stackvar).assign_coords(data.coords)
        else:
            #print ('2D')
            ds = stack[0].assign_coords(data.coords)
        del stack

        # it works faster when we prevent small output chunks
        chunksizes = {'y': self.chunksize, 'x': self.chunksize}
        if coarsen is not None:
            # coarse grid typically to square cells
            return ds.coarsen({'y': coarsen[0], 'x': coarsen[1]}, boundary='trim').mean().chunk(chunksizes)
        return ds.chunk(chunksizes)




# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_reframe_gmtsar import Stack_reframe_gmtsar
from .S1 import S1
from .PRM import PRM

class Stack_reframe(Stack_reframe_gmtsar):

    def _reframe_subswath(self, subswath, date, geometry, debug=False):
        """
        Estimate framed area using Sentinel-1 GCPs approximation.

        Parameters
        ----------
        subswath : int
            The subswath number.
        date : str
            The date of the scene.
        geometry: shapely.geometry of geopandas.GeoSeries or geopandas.GeoDataFrame
            Optional geometry covering required bursts to crop the area.
        debug : bool, optional
            Enable debug mode. Default is False.

        Returns
        -------
        pandas.DataFrame
            The updated DataFrame with the estimated framed area.

        Examples
        --------
        df = stack.reframe(1, '2023-05-20')
        """
        import geopandas as gpd
        import numpy as np
        #import shapely
        from shapely.geometry import Point, LineString, Polygon, MultiPolygon
        from shapely.ops import cascaded_union
        from datetime import datetime
        import os
        import warnings
        warnings.filterwarnings('ignore')

        # define line covering some bursts to crop them
        if isinstance(geometry, (gpd.GeoDataFrame, gpd.GeoSeries)):
            geometry = geometry.unary_union
        assert not geometry is None, f'ERROR: subswath {subswath} is not covered, you need to exclude it.'

        # convert to polygon when possible
        geometry = geometry.minimum_rotated_rectangle
        # it can be point or line or polygon
        if isinstance(geometry, Point):
            # create ~100m buffer around
            #geometry = geometry.buffer(1e-3)
            raise ValueError(f"Unsupported Point geometry. Unfortunately, GMTSAR tools cannot crop a scene to a single burst.")
        if isinstance(geometry, Polygon):
            rect = geometry.minimum_rotated_rectangle.exterior
            # define diagonal line
            diag1 = LineString([rect.coords[0], rect.coords[2]])
            diag2 = LineString([rect.coords[1], rect.coords[3]])
            if diag1.length <= diag2.length:
                geometry = diag1
            else:
                geometry = diag2
        if debug:
            print ('DEBUG: geometry', geometry)

        df = self.get_repeat(subswath, date)
        if debug:
            print('DEBUG: reframe scenes: ', len(df))
        stem = self.multistem_stem(subswath, df['datetime'].iloc[0])[1]
        #print ('stem', stem)

        old_filename = os.path.join(self.basedir, f'{stem}')
        #print ('old_filename', old_filename)

        self._make_s1a_tops(subswath, date, debug=debug)
        prm = PRM.from_file(old_filename+'.PRM')
        tmpazi_a = prm.SAT_llt2rat([geometry.coords[0][0],  geometry.coords[0][1],  0], precise=1, debug=debug)[1]
        tmpazi_b = prm.SAT_llt2rat([geometry.coords[-1][0], geometry.coords[-1][1], 0], precise=1, debug=debug)[1]
        tmpazi = min(tmpazi_a, tmpazi_b)
        if debug:
            print ('DEBUG: ','tmpazi', tmpazi)
        prm.shift_atime(tmpazi, inplace=True).update()
        azi_a = prm.SAT_llt2rat([geometry.coords[0][0], geometry.coords[0][1], 0], precise=1, debug=debug)[1] + tmpazi
        azi_b = prm.SAT_llt2rat([geometry.coords[-1][0], geometry.coords[-1][1], 0], precise=1, debug=debug)[1] + tmpazi
        # reorder boundaries for orbit
        azi1 = min(azi_a, azi_b)
        azi2 = max(azi_a, azi_b)
        if debug:
            print ('DEBUG: ','azi1', azi1, 'azi2', azi2)

        # Working on bursts covering $azi1 ($ll1) - $azi2 ($ll2)...
        #print ('_assemble_tops', subswath, date, azi1, azi2, debug)
        self._assemble_tops(subswath, date, azi1, azi2, debug=debug)

        # Parse new .xml to define new scene name
        # like to 's1b-iw3-slc-vv-20171117t145922-20171117t145944-008323-00ebab-006'
        filename = os.path.splitext(os.path.split(df['datapath'].iloc[0])[-1])[0]
        head1 = filename[:15]
        tail1 = filename[-17:]
        xml_header = S1.read_annotation(old_filename+'.xml')['product']['adsHeader']
        date_new = xml_header['startTime'][:10].replace('-','')
        t1 = xml_header['startTime'][11:19].replace(':','')
        t2 = xml_header['stopTime'][11:19].replace(':','')
        new_name = f'{head1}{date_new}t{t1}-{date_new}t{t2}-{tail1}'
        new_filename = os.path.join(self.basedir, new_name)
        #print ('new_filename', new_filename)

        # rename xml and tiff
        for ext in ['.tiff', '.xml']:
            if debug:
                print('DEBUG: rename', old_filename+ext, new_filename+ext)
            os.rename(old_filename+ext, new_filename+ext)

        # cleanup
        for fname in [old_filename+'.LED', old_filename+'.PRM']:
            if not os.path.exists(fname):
                continue
            if debug:
                print ('DEBUG: remove', fname)
            os.remove(fname)

        # update and return only one record
        out = df.head(1)
        #df['datetime'] = self.text2date(f'{date_new}t{t1}', False)
        out['datetime'] = datetime.strptime(f'{date_new}T{t1}', '%Y%m%dT%H%M%S')
        out['metapath'] = new_filename + '.xml'
        out['datapath'] = new_filename + '.tiff'
        # update approximate location
        out['geometry'] = cascaded_union([geom for multi_polygon in df.geometry for geom in multi_polygon.geoms
                                      if geom.intersects(geometry)])
        return out

    def compute_reframe(self, geometry=None, n_jobs=-1, **kwargs):
        """
        Reorder bursts from sequential scenes to cover the full orbit area or some bursts only.

        Parameters
        ----------
        geometry: shapely.geometry of geopandas.GeoSeries or geopandas.GeoDataFrame
            Optional geometry covering required bursts to crop the area.
        n_jobs : int, optional
            Number of parallel processing jobs. n_jobs=-1 means all the processor cores are used.

        Returns
        -------
        None

        Examples
        --------
        Without defined geometry the command is silently skipped:
        stack.reframe()
        
        Define a line partially covering two bursts:
        stack.reframe(geometry=LineString([Point(25.3, 35.0), Point(25, 35.2)]))
        
        Read the geometry from GeoJSON file and convert to WGS84 coordinates:
        AOI = gpd.GeoDataFrame().from_file('AOI.json').to_crs(4326)
        stack.reframe(geometry=AOI)
        
        TODO: Define a point on a selected burst (this option is not available now):
        stack.reframe(geometry=Point(25.3, 35))
        """
        from tqdm.auto import tqdm
        import joblib
        import pandas as pd

        dates = self.df.index.unique().values
        subswaths = self.get_subswaths()
        # approximate subswath geometries from GCP
        geometries = {subswath: self.df[self.df.subswath==subswath].geometry.unary_union for subswath in subswaths}

        # process all the scenes
        with self.tqdm_joblib(tqdm(desc='Reframing', total=len(dates)*len(subswaths))) as progress_bar:
            records = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(self._reframe_subswath)\
                            (subswath, date,
                            geometry.intersection(geometries[subswath]) if geometry is not None else geometries[subswath],
                            **kwargs) for date in dates for subswath in subswaths)

        self.df = pd.concat(records)
